import itertools
from typing import Sequence, Optional, Tuple, Dict
import numpy

from .base import Transformation
from .homography import Homography


__all__ = ["NumpyHomography"]


class NumpyHomography(
    Homography, registry_id=Homography.RegistryId("Homography", "Numpy")
):
    def __init__(self, *args, warp_options: Optional[Dict] = None, **kw) -> None:
        if warp_options is None:
            warp_options = dict()
        self._warp_options = warp_options
        super().__init__(*args, **kw)

    def apply_coordinates(self, coord: Sequence[numpy.ndarray]) -> numpy.ndarray:
        """
        :param coord: shape `(N, M)`
        :returns: shape `(N, M)`
        """
        return homography_transform_coordinates(self.active, coord)

    def apply_data(
        self,
        data: numpy.ndarray,
        offset: Optional[numpy.ndarray] = None,
        shape: Optional[numpy.ndarray] = None,
        cval=numpy.nan,
        interpolation_order: int = 1,
    ) -> numpy.ndarray:
        """
        :param data: shape `(N1, N2, ..., M1, M2, ...)` with `len((N1, N2, ...)) = N`
        :param offset: shape `(N,)`
        :param shape: shape `(N,) = [N1', N2', ...]`
        :param cval: missing value
        :param interpolation_order: order of interpolation: 0 is nearest neighbor, 1 is bilinear,...
        :returns: shape `(N1', N2', ..., M1, M2, ...)`
        """
        kw = dict(self._warp_options)
        if shape is not None:
            kw["shape"] = shape
        if offset is not None:
            kw["offset"] = offset
        if cval is not None:
            kw["cval"] = cval
        return homography_transform_data(self.passive, data, **kw)

    def __mul__(self, other: Transformation):
        if isinstance(other, NumpyHomography):
            return NumpyHomography(
                other.passive @ self.passive, warp_options=self._warp_options
            )
        else:
            ValueError("Only concatenation of same types allowed")


def homography_transform_coordinates(
    active: numpy.ndarray, coord: Sequence[numpy.ndarray]
) -> numpy.ndarray:
    """
    :param active: shape `(N+1, N+1)` or `(K, N+1, N+1)`
    :param coord: shape `(N, M)`
    :returns: shape `(N, M)` or `(K, N, M)`
    """
    N, M = coord.shape
    if active.ndim not in (2, 3) or active.shape[-2:] != (N + 1, N + 1):
        raise ValueError("matrix and coordinates dimensions do not match")
    coord = numpy.vstack(list(coord) + [numpy.ones(M)])
    hcoord = active.dot(coord)
    return hcoord[..., :-1, :] / hcoord[..., -1, :]


def homography_transform_bounding_box(
    active: numpy.ndarray, shape: Tuple[int]
) -> Tuple[numpy.ndarray, numpy.ndarray]:
    """Calculate the bounding box after transforming an image with shape `shape`.
    Returns the minimum and maximum coordinates of the bounding box.

    :param active: shape `(N+1, N+1)` or `(K, N+1, N+1)`
    :param shape: shape `(N,)`
    :returns: 2-tuple of arrays with shape `(N,)` or `(K, N)`
    """
    from_bounding_box = numpy.array(list(itertools.product(*[[0, n] for n in shape]))).T
    to_bounding_box = homography_transform_coordinates(active, from_bounding_box)
    return to_bounding_box.min(axis=-1), to_bounding_box.max(axis=-1)


def homography_transform_data(
    passive: numpy.ndarray,
    data: numpy.ndarray,
    offset: Optional[numpy.ndarray] = None,
    shape: Optional[numpy.ndarray] = None,
    cval=numpy.nan,
) -> numpy.ndarray:
    """The `offset` and `shape` determine the coordinates in the output frame
    at which to interpolate the data. By default these are the coordinates of
    the `data` in the input frame (i.e. the pixel indices).

    Only nearest-neighbor interpolation is supported.

    :param passive: shape `(N+1, N+1)`
    :param data: shape `(N1, N2, ..., M1, M2, ...)` with `len((N1, N2, ...)) = N`
    :param offset: shape `(N,)`
    :param shape: shape `(N,) = [N1', N2', ...]`
    :param cval: missing value
    :returns: shape `(N1', N2', ..., M1, M2, ...)`
    """
    N = passive.shape[0] - 1
    if passive.shape != (N + 1, N + 1):
        raise ValueError("requires a square matrix")
    if data.ndim < N:
        raise ValueError("passive and data dimensions do not match")

    if offset is not None:
        offset = numpy.asarray(offset)
        if offset.shape != (N,):
            raise ValueError("passive and offset dimensions do not match")
        passive = passive.copy()
        passive[:N, N] = offset

    shape_data = numpy.asarray(data.shape[:N])
    if shape is None:
        shape_out = shape_data.copy()
    else:
        shape_out = numpy.asarray(shape)
        if shape_out.shape != (N,):
            raise ValueError("passive and shape dimensions do not match")

    # Output coordinates
    xout = [numpy.arange(n, dtype=int) for n in shape_out]
    xout = numpy.meshgrid(*xout, indexing="ij")
    xout = numpy.array([x.flatten() for x in xout])  # (N, M) with M=N1*N2*...

    # Corresponding input coordinates
    xdata = numpy.round(homography_transform_coordinates(passive, xout)).astype(int)
    valid = (xdata >= 0) & (xdata < shape_data[:, None])
    valid = numpy.bitwise_and.reduce(valid, axis=0)

    # Nearest-neighbor interpolation
    out = numpy.full(shape_out, cval, dtype=data.dtype)
    if valid.size == 0:
        return out
    out[tuple(xout[:, valid])] = data[tuple(xdata[:, valid])]
    return out



import h5py
import numpy as np

from typing import Sequence, List, Tuple, Dict


def calculate_exclusions(string: str) -> Sequence[float]:
    exclusions = []
    if string is None:
        return exclusions
    slices = string.replace(" ", "").split(",")
    for slice in slices:
        try:
            if slice.split("-")[0] == slice:
                exclusions.append(float(slice + ".1"))
            else:
                split = slice.split("-")
                start = float(split[0] + ".1")
                stop = float(split[1] + ".1")
                exclusions.extend(np.arange(start, stop))
        except ValueError:
            raise ValueError(
                'The exclude syntax should be like "3, 12-15, 56, 57 - 58 "'
            )
    return exclusions


def get_stack(
    path: str,
    stack: str,
    exclusions: Sequence = list(),
    energy_counter: str = "Edcmund",
) -> Tuple[List[np.ndarray], List[float]]:
    images = list()
    energies = list()
    with h5py.File(path) as f:
        title = f["2.1"]["title"].asstr()[()]
        title = title.split()
        ysize, xsize = int(title[4]), int(title[-2]) + 1
        length = xsize * ysize

        # Iterating over f.keys() is not possible as the keys are in lexicographic order
        for index in np.arange(1.1, len(f.keys()) + 1, 1.0):
            if index in exclusions:
                continue
            index = str(index)
            if (
                "measurement" in f[index].keys()
                and stack in f[index]["measurement"].keys()
            ):
                flat = f[index]["measurement"][stack]
                if np.all(np.array(flat) == 0):
                    break
                if flat.size == length and np.all(np.isfinite(flat)):
                    image = np.reshape(flat, (xsize, ysize))
                    images.append(image)
                    energies.append(
                        f[index]["instrument"]["positioners"][energy_counter][()]
                    )

    if len(images) == 0:
        raise ValueError("With the given configuration, no images were collected")
    return images, energies


def get_stack_dict(
    path: str, stack: str, exclusions: Sequence = list()
) -> Tuple[List[np.ndarray], List[float]]:
    images = dict()
    with h5py.File(path) as f:
        title = f["2.1"]["title"].asstr()[()]
        title = title.split()
        ysize, xsize = int(title[4]), int(title[-2]) + 1
        length = xsize * ysize

        # Iterating over f.keys() is not possible as the keys are in lexicographic order
        for energy in np.arange(1.1, len(f.keys()) + 1, 1.0):
            if energy in exclusions:
                continue
            energy = str(energy)
            if (
                "measurement" in f[energy].keys()
                and stack in f[energy]["measurement"].keys()
            ):
                flat = f[energy]["measurement"][stack]
                if np.all(np.array(flat) == 0):
                    break
                if flat.size == length and np.all(np.isfinite(flat)):
                    image = np.reshape(flat, (xsize, ysize))
                    images[energy] = image

    if len(images) == 0:
        raise ValueError("With the given configuration, no images were collected")
    return images


def get_all_stacks(
    path: str, stacks: List[str], exclusions: Sequence = list()
) -> Dict[str, Dict[str, np.ndarray]]:
    imagestacks = dict()
    with h5py.File(path) as f:
        title = f["2.1"]["title"].asstr()[()]
        title = title.split()
        ysize, xsize = int(title[4]), int(title[-2]) + 1
        length = xsize * ysize

        # Iterating over f.keys() is not possible as the keys are in lexicographic order
        bad_stacks = []

        for energy in np.arange(2.1, len(f.keys()) + 1, 1.0):
            images = dict()
            if energy in exclusions:
                continue
            energy = str(energy)
            try:
                scan = f[energy]["measurement"]
            except KeyError:
                continue
            if all(
                [scan[stack].size != length for stack in stacks if stack in scan.keys()]
            ):
                continue
            for stack in stacks:
                if stack not in scan.keys():
                    bad_stacks.append(stack)
                    continue
                image = scan[stack]
                if np.all(np.array(image) == 0) or image.size != length:
                    bad_stacks.append(stack)
                    continue
                if not np.all(np.isfinite(image)):
                    bad_stacks.append(stack)
                    continue
                image = np.reshape(image, (xsize, ysize))
                images[stack] = image

            stacks = [stack for stack in stacks if stack not in bad_stacks]

            # if all counters only have valid values, we can add this energy

            imagestacks[energy] = images
        # for every energy remove all stacks that are faulty at some energy
        [
            [scan.pop(bad) for bad in bad_stacks if bad in scan.keys()]
            for scan in imagestacks.values()
        ]

    if len(imagestacks) == 0:
        raise ValueError("With the given configuration, no images were collected")
    return imagestacks



import pytest
from .data import data_for_registration
import numpy
from ..transformation.scikitimage_backend import SciKitImageHomography


@pytest.mark.parametrize(
    "transfo_type", ["translation", "rigid", "similarity", "affine"]
)
def test_images(transfo_type):
    data_for_registration.images(transfo_type, plot=0)


@pytest.mark.parametrize("transfo_type", ["rigid", "similarity", "affine"])
def test_transformations(transfo_type):
    images, active, passive = data_for_registration.images(
        transfo_type, nimages=5, name="gravel", plot=0
    )
    back = [
        SciKitImageHomography(passive=mat, warp_options={"order": 3}) for mat in passive
    ]
    results = [hom.apply_data(images[i]) for i, hom in enumerate(back)]

    diffs = [(images[0] - res) for res in results]

    numpy.testing.assert_allclose(numpy.nanmax(diffs, axis=(1, 2)), 0, atol=0.2)



NAME = "Registration"

DESCRIPTION = "Data registration"

LONG_DESCRIPTION = "Data registration"

BACKGROUND = "light-blue"

WIDGET_HELP_PATH = (
    # Development documentation (make htmlhelp in ./doc)
    (
        "/home/sireiche/ewoksndreg/build/sphinx/html/index.html",
        None,
    ),
    # Documentation included in wheel
    # ("{}/help/ewoksndreg/index.html".format(sysconfig.get_path("data")), None),
    # Online documentation url
    ("https://ewoksndreg.readthedocs.io/en/latest/", ""),
)



import numpy
from .base import Features


__all__ = ["NumpyKeypointFeatures"]


class NumpyKeypointFeatures(
    Features, registry_id=Features.RegistryId("Keypoints", "Numpy")
):
    def __init__(self, keypoints: numpy.ndarray) -> None:
        nfeatures, ndim = keypoints.shape
        self._coordinates = keypoints.T
        self._ndim = ndim
        self._nfeatures = nfeatures
        super().__init__()

    def __getitem__(self, idx):
        return type(self)(self.keypoints[idx])

    @property
    def keypoints(self) -> numpy.ndarray:
        return self._keypoints

    @property
    def coordinates(self) -> numpy.ndarray:
        return self._coordinates

    @property
    def ndim(self) -> int:
        return self._ndim

    @property
    def nfeatures(self) -> int:
        return self._nfeatures



from contextlib import contextmanager
import numpy

from silx.gui.plot import Plot2D
from silx.gui.widgets.FrameBrowser import HorizontalSliderWithBrowser

from ewoksorange.bindings import OWEwoksWidgetOneThread
from ewoksorange.bindings import ow_build_opts
from ewoksorange.gui.parameterform import ParameterForm

from ewoksndreg.tasks import Reg2DFeatures
from ewoksndreg.features import registration
from ewoksndreg.transformation.types import TransformationType
from ewoksndreg.io.input_stack import input_context


__all__ = ["OWReg2DFeatures"]


class OWReg2DFeatures(
    OWEwoksWidgetOneThread, **ow_build_opts, ewokstaskclass=Reg2DFeatures
):
    name = "2D Feature-Based Registration"
    description = "Calculate transformations of a stack of images based on features"
    icon = "icons/2d_features.svg"

    def __init__(self):
        super().__init__()
        self._space = 0.05
        self._init_control_area()
        self._init_main_area()

    def _init_control_area(self):
        super()._init_control_area()
        self._default_inputs_form = ParameterForm(parent=self.controlArea)
        values = self.get_default_input_values(include_missing=True)

        options = {
            "imagestack": {"value_for_type": "", "select": "h5dataset"},
            "inputs_are_stacks": {"value_for_type": False},
            "reference": {"value_for_type": 0},
            "transformation_type": {
                "value_for_type": list(TransformationType.__members__),
                "serialize": str,
            },
            "detector": {
                "value_for_type": registration.FeatureDetector.get_subclass_ids(),
                "serialize": str,
            },
            "matcher": {
                "value_for_type": registration.FeatureMatching.get_subclass_ids(),
                "serialize": str,
            },
            "mapper": {
                "value_for_type": registration.FeatureMapping.get_subclass_ids(),
                "serialize": str,
            },
        }

        for name, kw in options.items():
            self._default_inputs_form.addParameter(
                name,
                value=values[name],
                value_change_callback=self._default_inputs_changed,
                **kw,
            )

    def _default_inputs_changed(self):
        self.update_default_inputs(**self._default_inputs_form.get_parameter_values())
        self._update_data()

    def handleNewSignals(self) -> None:
        self._update_input_data()
        super().handleNewSignals()

    def task_output_changed(self):
        self._update_data()

    def _init_main_area(self):
        super()._init_main_area()
        layout = self._get_main_layout()
        self._plot = Plot2D(parent=self.mainArea)
        layout.addWidget(self._plot)
        self._slider = HorizontalSliderWithBrowser(parent=self.mainArea)
        layout.addWidget(self._slider)
        self._slider.valueChanged[int].connect(self._select_image)
        self._update_data()

    @contextmanager
    def _input_context(self, images=None):
        if images is not None:
            yield images
            return
        try:
            binit = True
            with input_context(self.get_task_input_value("imagestack")) as images:
                binit = False
                yield images
        except TypeError:
            if binit:
                yield list()

    def _update_data(self):
        with self._input_context() as images:
            self._slider.setMaximum(max(len(images) - 1, 0))
            self._select_image(self._slider.value(), images=images)

    def _update_input_data(self):
        dynamic = self.get_dynamic_input_names()
        for name in self.get_input_names():
            self._default_inputs_form.set_parameter_enabled(name, name not in dynamic)

    def _select_image(self, select, images=None):
        with self._input_context(images=images) as images:
            if not images:
                return
            reference = self.get_task_input_value("reference", 0)
            refimg = images[reference]
            selimg = images[select]

            space = max(int(refimg.shape[1] * self._space), 1)
            spaceimg = numpy.full((refimg.shape[0], space), numpy.nan)
            off = refimg.shape[1] + space

            self._plot.addImage(
                numpy.hstack([refimg, spaceimg, selimg]), legend="images"
            )

            transformations = self.get_task_output_value("transformations")
            if not transformations:
                return
            features = self.get_task_output_value("features")
            matches = self.get_task_output_value("matches")

            def legend_generator():
                i = 0
                while True:
                    yield f"c{i}"
                    i += 1

            legend = legend_generator()

            self._plot.remove(kind="curve")

            yref, xref = features[reference].coordinates
            ysel, xsel = features[select].coordinates
            xsel = xsel + off

            self._plot.addCurve(
                xref,
                yref,
                legend=next(legend),
                symbol="+",
                linestyle=" ",
                color="#00FF00",
            )
            self._plot.addCurve(
                xsel,
                ysel,
                legend=next(legend),
                symbol="+",
                linestyle=" ",
                color="#00FF00",
            )

            reffeatures, selfeatures = matches[select]
            if reffeatures is None:
                reffeatures = features[reference]
                selfeatures = reffeatures
            yref, xref = reffeatures.coordinates
            ysel, xsel = selfeatures.coordinates
            xsel = xsel + off

            for x0, y0, x1, y1 in zip(xref, yref, xsel, ysel):
                self._plot.addCurve(
                    [x0, x1],
                    [y0, y1],
                    legend=next(legend),
                    color="#FF0000",
                    linestyle="-",
                )



import torch
import torch.nn.functional as F


def nmi_loss(
    source: torch.Tensor, target: torch.Tensor, bandwidth: int = 0.01, reduction="none"
):
    """
    Calculates the Normalized Mutual Information between batch source and batch target

    param source: Batch of Tensors
    param target: Batch of Tensors with same size as source
    param bandwidth: The size of the kernel function, determines the sample rate of the histogramm(default: 0.01)
    param reduction: If None: returns Tensor of NMI for every element in batch. If Sum or Mean: Applies operation to the resulting tensor

    """
    eps = 1e-10
    m = source.flatten(start_dim=1)
    t = target.flatten(start_dim=1)
    batch = m.size()[0]
    bins = round(0.5 / bandwidth)

    samples = torch.linspace(0, 1, bins)
    # samples has size [bins]
    samples = samples.unsqueeze(0).expand((batch, -1))
    # m has size [N,items], samples has size [N, bins]
    m_outer = m[..., None] - samples[:, None, ...]
    m_outer = quartic_kernel(m_outer / bandwidth)

    t_outer = t[:, None, ...] - samples[..., None]
    t_outer = quartic_kernel(t_outer / bandwidth)
    # m_outer and t_outer have shape [N,bins, items] and [N,items,bins]
    # result has dimensions[N,bins,bins]

    joint_hgram = torch.matmul(t_outer, m_outer)
    joint_hgram = F.normalize(joint_hgram, p=1, dim=(1, 2))
    m_hgram = torch.sum(joint_hgram, dim=1)
    t_hgram = torch.sum(joint_hgram, dim=2)

    m_hgram[m_hgram <= eps] = eps
    t_hgram[t_hgram <= eps] = eps
    joint_hgram[joint_hgram <= eps] = eps

    m_ent = (m_hgram * torch.log2(m_hgram)).nansum(dim=1)
    t_ent = (t_hgram * torch.log2(t_hgram)).nansum(dim=1)
    joint_ent = (joint_hgram * torch.log2(joint_hgram)).nansum(dim=(1, 2))

    mut_inf = -((m_ent + t_ent) / joint_ent)
    if reduction == "none":
        return mut_inf
    elif reduction == "mean":
        return mut_inf.mean()
    elif reduction == "sum":
        return mut_inf.sum()
    else:
        raise ValueError(f"reduction must either be none, mean or sum. Got {reduction}")


def quartic_kernel(data: torch.Tensor) -> torch.Tensor:
    result = torch.zeros_like(data)
    result[torch.logical_and(data < 1, data > -1)] = (
        15 / 16 * (1 - data[torch.logical_and(data < 1, data > -1)] ** 2) ** 2
    )
    return result


def block_kernel(data: torch.Tensor) -> torch.Tensor:
    result = torch.zeros_like(data)
    result[torch.logical_and(data < 1, data > -1)] = 1
    return result



"""Ewoks workflow tasks"""

from .example2d_stack import *  # noqa F401
from .reg2d_features import *  # noqa F401
from .reg2d_intensities import *  # noqa F401
from .reg2d_transform import *  # noqa F401
from .fluo_stack import *  # noqa F401
from .reg2d_multistack import *  # noqa F401
from .select_slice import *  # noqa F401
from .id16b_file_read import *  # noqa F401



"""Intensity based registration"""

from .base import *  # noqa F401
from .numpy_backend import *  # noqa F401

try:
    from .simpleitk_backend import *  # noqa F401
except ImportError:
    pass


try:
    from .scikitimage_backend import *  # noqa F401
except ImportError:
    pass

try:
    from .kornia_backend import *  # noqa F401
except ImportError:
    pass



from typing import Optional, Mapping
import numpy
from silx.opencl.common import ocl
from silx.opencl import sift
from .base import FeatureDetector
from ..features import SilxDescriptorFeatures

if ocl is None:
    raise ImportError("pyopencl missing")

__all__ = ["SilxSiftFeatureDetector"]


class SilxSiftFeatureDetector(
    FeatureDetector, registry_id=FeatureDetector.RegistryId("Sift", "Silx")
):
    def __init__(self, feature_options: Optional[Mapping] = None, **kw) -> None:
        if feature_options is None:
            feature_options = dict()
        self._feature_options = feature_options
        self._feature_detector = None
        super().__init__(**kw)

    def find(self, image: numpy.ndarray) -> SilxDescriptorFeatures:
        if (
            self._feature_detector is None
            or self._feature_detector.dtype is not image.dtype
            or self._feature_detector.shape != image.shape
        ):
            self._feature_detector = sift.SiftPlan(
                dtype=image.dtype, shape=image.shape, **self._feature_options
            )
        silx_features = self._feature_detector.keypoints(image)

        if self._mask is not None:
            kpx = numpy.round(silx_features.x).astype(numpy.int32)
            kpy = numpy.round(silx_features.y).astype(numpy.int32)
            masked = self._mask[(kpy, kpx)].astype(bool)
            silx_features = silx_features[masked]

        return SilxDescriptorFeatures(silx_features)



from ewokscore.task import Task
from ..io.input_stack import input_context
from ..io.output_stack import output_context
from ..transformation import apply_transformations, Transformation
from ..transformation import SciKitImageHomography

from os.path import join
from silx.io.url import DataUrl

import numpy as np
from itertools import cycle
from typing import Union, Sequence

__all__ = ["Reg2DTransform"]


class Reg2DTransform(
    Task,
    input_names=["imagestack", "transformations"],
    optional_input_names=[
        "url",
        "inputs_are_stacks",
        "crop",
        "interpolation_order",
        "stack_names",
    ],
    output_names=["imagestack"],
):
    """Apply transformations calculated from image registration to images."""

    def run(self):
        url = self.get_input_value("url", None)
        crop = self.get_input_value("crop", False)
        interpolation_order = self.get_input_value("interpolation_order", 1)
        stack_names = self.get_input_value("stack_names", None)

        # the transformations are either type Transformation or np.ndarray(homography matrices)
        transfos = as_transformations(self.inputs.transformations)

        with input_context(
            self.inputs.imagestack,
            inputs_are_stacks=self.get_input_value("inputs_are_stacks", None),
        ) as istack:
            if istack.ndim == 3:
                if isinstance(transfos[0], np.ndarray):
                    transfos = [SciKitImageHomography(tr) for tr in transfos]
                if len(transfos) != len(istack):
                    raise ValueError(
                        "Amount of transformations and images doesn't match"
                    )
                with output_context(url=url) as ostack:
                    apply_transformations(
                        istack,
                        ostack,
                        transfos,
                        crop=crop,
                        interpolation_order=interpolation_order,
                    )
                    if url:
                        self.outputs.imagestack = url
                    else:
                        self.outputs.imagestack = ostack.data
                return

            if stack_names is None:
                stack_names = [str(i) for i in range(len(istack))]
            stack_names = [name + "_aligned" for name in stack_names]
            if isinstance(transfos[0], (np.ndarray, Transformation)):
                transfos = [transfos]
            else:
                if crop:
                    raise ValueError(
                        "Cropping when applying multiple different lists not supported because of differing image sizes"
                    )

            # we have a list of list of transformations and a list of stacks
            if len(transfos[0]) != len(istack[0]):
                raise ValueError(
                    "Amount of transformations and images per stack doesn't match"
                )
            if not (len(transfos) == 1 or len(transfos) == len(istack)):
                raise ValueError(
                    "Amount of imagestacks and transformationstacks doesn't match"
                )
            data = []
            for stack, transfo, name in zip(istack, cycle(transfos), stack_names):
                if url:
                    # append `name` to the url
                    dataurl = DataUrl(url)
                    dpath = dataurl.data_path()
                    if dpath is None:
                        dpath = "/"
                    dpath = join(dpath, name)
                    full_url = DataUrl(
                        file_path=dataurl.file_path(), data_path=dpath
                    ).path()
                else:
                    full_url = None
                with output_context(url=full_url) as ostack:
                    apply_transformations(
                        stack,
                        ostack,
                        transfo,
                        crop=crop,
                        interpolation_order=interpolation_order,
                    )
                    if full_url:
                        data.append(full_url)
                    else:
                        data.append(np.asarray(ostack.data))

            self.outputs.imagestack = data


def as_transformations(
    transformations: Union[
        Sequence[np.ndarray],
        Sequence[Transformation],
        Sequence[Sequence[Transformation]],
        Sequence[Sequence[np.ndarray]],
    ]
) -> Union[Sequence[Transformation], Sequence[Sequence[Transformation]]]:
    if isinstance(transformations[0], np.ndarray):
        return [SciKitImageHomography(transfo) for transfo in transformations]

    if isinstance(transformations[0], Sequence) and isinstance(
        transformations[0][0], np.ndarray
    ):
        return [[SciKitImageHomography(t) for t in seq] for seq in transformations]

    return transformations



import numpy
from .fit2d import fitgaussian as fitgaussian2d
from .fit1d import fitgaussian as fitgaussian1d


def fmax(data):
    if data.size in data.shape:
        return numpy.nanargmax(data)
    else:
        return numpy.array(numpy.unravel_index(numpy.nanargmax(data), data.shape))


def fmin(data):
    if data.size in data.shape:
        return numpy.nanargmin(data)
    else:
        return numpy.array(numpy.unravel_index(numpy.nanargmin(data), data.shape))


def fcentroid(data):
    return foptimize(data, _centroid)


def fgaussmax(data):
    return foptimize(data, _fit)


def foptimize(data, proc, threshold=0.9):
    shift = fmax(data)
    thres = threshold * numpy.nanmax(data)

    if data.size in data.shape:
        shifta = shift
        shiftb = shift
        while (data.flat[shifta] > thres) and (data.flat[shiftb] > thres):
            shifta -= 1
            shiftb += 1
            if shifta < 0 or shiftb >= data.size:
                shifta += 1
                shiftb -= 1
                break
        if shifta != shiftb:
            shift = proc(data.flat[shifta : shiftb + 1]) + shifta
    else:
        off = 0
        s = data.shape
        while (
            data[
                shift[0] - off : shift[0] + off + 1, shift[1] - off : shift[1] + off + 1
            ]
            < thres
        ).sum(dtype=int) == 0:
            off += 1
            if (
                shift[0] < off
                or shift[1] < off
                or shift[0] + off >= s[0]
                or shift[1] + off >= s[1]
            ):
                off -= 1
                break

        if off != 0:
            shift = (
                shift
                + proc(
                    data[
                        shift[0] - off : shift[0] + off + 1,
                        shift[1] - off : shift[1] + off + 1,
                    ]
                )
                - off
            )

    return shift


def _centroid(data):
    if data.size in data.shape:
        x = numpy.arange(data.size).reshape(data.shape)
        return (x * data).sum() / data.sum()
    else:
        ny, nx = numpy.shape(data)
        y, x = numpy.indices((ny, nx))
        cx = numpy.sum(x * data) / numpy.sum(data)
        cy = numpy.sum(y * data) / numpy.sum(data)
        return numpy.array((cx, cy))


def _fit(data):
    if data.size in data.shape:
        x = numpy.arange(data.size)
        p, success = fitgaussian1d(x, data)
        if success:
            ret = p[0]
    else:
        y, x = numpy.indices(data.shape)
        p, success = fitgaussian2d(x, y, data)
        if success:
            ret = p[[0, 1]]
    if success:
        return ret
    else:
        return fmax(data)



from ewokscore.task import Task

from ..io.input_stack import InputStackNumpy
from ..transformation.numpy_backend import NumpyHomography
import numpy

try:
    from skimage.util import random_noise
except ImportError:
    random_noise = None

try:
    from ..transformation.simpleitk_backend import SimpleITKTransformation

    simpleitk = True
except ImportError:
    simpleitk = False
from ..tests.data import data_for_registration


__all__ = ["Example2DStack"]


class Example2DStack(
    Task,
    input_names=["name", "transformation_type"],
    optional_input_names=["shape", "nimages", "add_noise_dim"],
    output_names=["imagestack", "transformations"],
):
    """Generate example images with successive transformations to test registration methods."""

    def run(self):
        values = self.get_input_values()
        transformation_type = values.pop("transformation_type")
        noisy = values.pop("add_noise_dim", False)
        istack, _, passive = data_for_registration.images(transformation_type, **values)

        if random_noise and noisy:
            n = 10
            data = [
                random_noise(numpy.array(istack), mode="s&p", amount=0.08 * i)
                for i in range(n)
            ]
            shuffle = numpy.random.permutation(n)
            data = [data[i] for i in shuffle]
            self.outputs.imagestack = InputStackNumpy(data, False)
        else:
            self.outputs.imagestack = istack

        if transformation_type in ["displacement_field", "bspline"]:
            if not simpleitk:
                raise ValueError(
                    "displacement field transforms cannot be generated without SimpleITK"
                )
            self.outputs.transformations = [
                SimpleITKTransformation(displacement_field=d) for d in passive
            ]
        else:
            self.outputs.transformations = [NumpyHomography(M) for M in passive]



import fabio
from ..tests.data.read import getfiles
from ..io.output_stack import output_context
from ewokscore import Task


class ID16BFileRead(
    Task,
    input_names=["scandir", "recsuffix"],
    optional_input_names=["url", "normalize"],
    output_names=["imagestack"],
):
    def run(self):
        scandir = self.get_input_value("scandir", None)
        recsuffix = self.get_input_value("recsuffix", None)
        url = self.get_input_value("url", None)
        normalize = self.get_input_value("normalize", False)
        energies, files = getfiles(scandir, recsuffix)
        with output_context(url=url) as ostack:
            for filename, energy in zip(files, energies):
                with fabio.open(filename) as f:
                    image = f.data
                    if normalize:
                        image = (image - image.min()) / (image.max() - image.min())
                    ostack.add_point(image)
            if url is not None:
                self.outputs.imagestack = url
            else:
                self.outputs.imagestack = ostack.data



from typing import Optional, List, Type, NamedTuple, Union, Tuple


class RegistryId(NamedTuple):
    name: str
    backend: str

    def __str__(self):
        return f"{self.name}-{self.backend}"

    @classmethod
    def factory(self, value: Union["RegistryId", str, Tuple]):
        if isinstance(value, RegistryId):
            return value
        if isinstance(value, str):
            return RegistryId(*value.split("-"))
        return RegistryId(*value)


class Registered:
    _SUBCLASS_REGISTRY = None

    RegistryId = RegistryId

    def __init_subclass__(
        cls,
        register: bool = True,
        registry_id: Union["RegistryId", str, Tuple, None] = None,
        **kwargs,
    ):
        super().__init_subclass__(**kwargs)

        # Ensures that not all subclasses share the same registry
        if cls._SUBCLASS_REGISTRY is None:
            cls._SUBCLASS_REGISTRY = dict()

        if not register:
            cls.__registry_id = None
            return

        # Register the subclass
        if registry_id is None:
            raise RuntimeError(f"Class {repr(cls)} is missing a `registry_id`")
        registry_id = RegistryId.factory(registry_id)
        ecls = cls._SUBCLASS_REGISTRY.get(registry_id)
        if ecls is not None:
            raise RuntimeError(
                f"Registry name {registry_id} is already taken by {repr(ecls)}"
            )
        cls.__registry_id = registry_id
        cls._SUBCLASS_REGISTRY[registry_id] = cls

    @classmethod
    def get_subclass_id(cls) -> Optional[RegistryId]:
        return cls.__registry_id

    @classmethod
    def get_subclass_ids(cls) -> List[RegistryId]:
        return list(cls._SUBCLASS_REGISTRY.keys())

    @classmethod
    def get_subclass_items(cls) -> List[RegistryId]:
        return list(cls._SUBCLASS_REGISTRY.items())

    @classmethod
    def get_subclasses(cls) -> List[Type["Registered"]]:
        return list(cls._SUBCLASS_REGISTRY.values())

    @classmethod
    def get_subclass(cls, registry_id: RegistryId) -> Type["Registered"]:
        registry_id = RegistryId.factory(registry_id)
        try:
            return cls._SUBCLASS_REGISTRY[registry_id]
        except KeyError:
            s = ", ".join(list(map(str, cls._SUBCLASS_REGISTRY)))
            raise RuntimeError(
                f"No class with registry name {registry_id} found. Available classes are {s}"
            )



"""Calculate active transformation between two sets of coordinates"""

from typing import Sequence, Callable, Dict
import numpy
from ...transformation import lstsq
from ...transformation import TransformationType


_METHODS: Dict[
    TransformationType,
    Callable[[Sequence[numpy.ndarray], Sequence[numpy.ndarray]], numpy.ndarray],
] = {
    TransformationType.identity: lstsq.calc_identity,
    TransformationType.translation: lstsq.calc_translation,
    TransformationType.rigid: lstsq.calc_rigid,
    TransformationType.similarity: lstsq.calc_similarity,
    TransformationType.affine: lstsq.calc_affine,
    TransformationType.projective: lstsq.calc_projective,
}


def get_lstsq_solver(
    transfo_type: TransformationType,
) -> Callable[[Sequence[numpy.ndarray], Sequence[numpy.ndarray]], numpy.ndarray]:
    try:
        return _METHODS[transfo_type]
    except KeyError:
        raise ValueError(f"No least-squares solver found for {transfo_type}") from None



import SimpleITK as sitk
import numpy
from typing import Optional
from typing_extensions import Literal
from .base import IntensityMapping
from ..transformation.types import TransformationType
from ..transformation.simpleitk_backend import SimpleITKTransformation
from .types import SitkMetricType, SitkOptimizerType


class sitkOptimizerIntesityMapping(
    IntensityMapping,
    registry_id=IntensityMapping.RegistryId("Optimization", "SimpleITK"),
):
    SUPPORTED_TRANSFORMATIONS = [
        "translation",
        "rigid",
        "similarity",
        "affine",
        "bspline",
        "displacement_field",
    ]

    def __init__(
        self,
        transfo_type: TransformationType,
        metric: SitkMetricType = SitkMetricType.MeanSquares,
        optimizer: SitkOptimizerType = SitkOptimizerType.LBFGSB,
        order: int = 1,
        mask: Optional[numpy.ndarray] = None,
        pyramid_levels: int = 1,
        mesh_size: Optional[tuple] = None,
        sampling: Optional[Literal["random", "regular"]] = None,
        sampling_percentage: int = 0.5,
        **kw,
    ) -> None:
        self._registration_method = sitk.ImageRegistrationMethod()
        self._metric = SitkMetricType(metric)
        self._optimizer = SitkOptimizerType(optimizer)
        self._mask = mask
        self._mesh_size = mesh_size
        self._levels = pyramid_levels

        if order == 0:
            self._registration_method.SetInterpolator(sitk.sitkNearestNeighbor)
        elif order == 1:
            self._registration_method.SetInterpolator(sitk.sitkLinear)
        elif order == 2:
            self._registration_method.SetInterpolator(sitk.sitkBSpline1)
        elif order == 3:
            self._registration_method.SetInterpolator(sitk.sitkBSpline2)
        elif order == 4:
            self._registration_method.SetInterpolator(sitk.sitkBSpline3)
        elif order == 5:
            self._registration_method.SetInterpolator(sitk.sitkBSpline4)

        self._order = order

        super().__init__(transfo_type, **kw)

        # set the metric
        if self.type == self.type.displacement_field:
            self._registration_method.SetMetricAsDemons(0.00001)
        elif self._metric == self._metric.CrossCorrelation:
            self._registration_method.SetMetricAsCorrelation()
        elif self._metric == self._metric.ANTSNeighborhoodCorrelation:
            self._registration_method.SetMetricAsANTSNeighborhoodCorrelation()
        elif self._metric == self._metric.JointHistogramMutualInformation:
            self._registration_method.SetMetricAsJointHistogramMutualInformation()
        elif self._metric == self._metric.MattesMutualInformation:
            self._registration_method.SetMetricAsMattesMutualInformation()
        elif self._metric == self._metric.MeanSquares:
            self._registration_method.SetMetricAsMeanSquares()
        else:
            raise ValueError(
                f"metric {self._metric} not supported by SimpleITK, choose from {list(SitkMetricType.__members__)}"
            )

        # set the optimizer
        if self._optimizer == self._optimizer.Exhaustive:
            self._registration_method.SetOptimizerAsExhaustive((10, 10))
        elif self._optimizer == self._optimizer.Powell:
            self._registration_method.SetOptimizerAsPowell()
        elif self._optimizer == self._optimizer.Evolutionary:
            self._registration_method.SetOptimizerAsOnePlusOneEvolutionary()
        elif self._optimizer == self._optimizer.LBFGSB:
            self._registration_method.SetOptimizerAsLBFGSB()
        elif self._optimizer == self._optimizer.GradientDescent:
            self._registration_method.SetOptimizerAsGradientDescent(
                learningRate=1.0,
                numberOfIterations=300,
                convergenceMinimumValue=1e-6,
                convergenceWindowSize=10,
            )
        elif self._optimizer == self._optimizer.CGLineSearch:
            self._registration_method.SetOptimizerAsConjugateGradientLineSearch(
                2.0, 300
            )
        elif self._optimizer == self._optimizer.GradientDescentLineSearch:
            self._registration_method.SetOptimizerAsGradientDescentLineSearch(2.0, 300)
        else:
            raise ValueError(
                f"optimizer {self._optimizer} not supported by SimpleITK, choose from {[str(item) for item in list(SitkOptimizerType)]}"
            )

        # set sampling strategy
        if sampling == "random":
            self._registration_method.SetMetricSamplingStrategy(
                self._registration_method.RANDOM
            )
            self._registration_method.SetMetricSamplingPercentage(sampling_percentage)
        elif sampling == "regular":
            self._registration_method.SetMetricSamplingStrategy(
                self._registration_method.REGULAR
            )
            self._registration_method.SetMetricSamplingPercentage(sampling_percentage)

    def identity(self, dimension: int = 2) -> SimpleITKTransformation:
        return SimpleITKTransformation(
            transfo_type="identity", passive=numpy.identity(dimension + 1)
        )

    def calculate(
        self,
        from_image: numpy.ndarray,
        to_image: numpy.ndarray,
    ) -> SimpleITKTransformation:
        if self._mask is not None and numpy.any(self._mask):
            self._mask = self._mask.astype(bool)
            self._registration_method.SetMetricMovingMask(
                sitk.GetImageFromArray(self._mask.astype("uint8"))
            )
            self._registration_method.SetMetricFixedMask(
                sitk.GetImageFromArray(self._mask.astype("uint8"))
            )

        from_imageitk = sitk.GetImageFromArray(from_image)
        to_imageitk = sitk.GetImageFromArray(to_image)

        dim = from_imageitk.GetDimension()

        if self.type == self.type.translation:
            tx = sitk.TranslationTransform(dim, dim * [0])
        elif self._type in ["rigid", "similarity", "affine"]:
            if self.type == self.type.rigid and dim == 2:
                tf = sitk.Euler2DTransform()
            elif self.type == self.type.rigid and dim == 3:
                tf = sitk.Euler3DTransform()
            elif self.type == self.type.similarity and dim == 2:
                tf = sitk.Similarity2DTransform()
            elif self.type == self.type.similarity and dim == 3:
                tf = sitk.Similarity3DTransform()
            elif self.type == self.type.affine:
                tf = sitk.AffineTransform(dim)
            tx = sitk.CenteredTransformInitializer(
                to_imageitk,
                from_imageitk,
                tf,
                sitk.CenteredTransformInitializerFilter.GEOMETRY,
            )
        elif self.type == self.type.bspline:
            # Use given mesh size or put control point every 20 pixels
            if self._mesh_size:
                tx = sitk.BSplineTransformInitializer(from_imageitk, self._mesh_size)
            else:
                tx = sitk.BSplineTransformInitializer(
                    from_imageitk, [int(size / 20) for size in from_imageitk.GetSize()]
                )
        elif self.type == self.type.displacement_field:
            img = sitk.GetImageFromArray(numpy.zeros((*from_image.shape, dim)), True)
            tx = sitk.DisplacementFieldTransform(img)
            tx.SetSmoothingGaussianOnUpdate(
                varianceForUpdateField=0.0, varianceForTotalField=2.0
            )
        else:
            raise ValueError(f"{self._type} is not supported by SimpleITK")

        self._registration_method.SetInitialTransform(tx)

        if self._optimizer not in ["Powell", "GradientDescent"]:
            factors = [int(2**x) for x in numpy.arange(self._levels - 1, -1, -1)]
            self._registration_method.SetShrinkFactorsPerLevel(shrinkFactors=factors)
            self._registration_method.SetSmoothingSigmasPerLevel(
                smoothingSigmas=numpy.arange(
                    self._levels - 1, -1, -1, dtype=numpy.float64
                )
            )
        if self._optimizer != self._optimizer.LBFGSB:
            self._registration_method.SetOptimizerScalesFromPhysicalShift()
        result = self._registration_method.Execute(to_imageitk, from_imageitk)

        return SimpleITKTransformation(
            transfo_type=self._type,
            transformation=result,
        )



import numpy
from typing import Optional
from .base import IntensityMapping
from ..transformation.types import TransformationType
from ..transformation.scikitimage_backend import SciKitImageHomography
from ..math.fft import fft2, fftshift

from packaging.version import Version
import skimage
from skimage.transform import warp_polar, SimilarityTransform
from skimage.registration import phase_cross_correlation as _phase_cross_correlation


def phase_cross_correlation(img1, img2, **kw) -> numpy.ndarray:
    version = Version(skimage.__version__)

    if version >= Version("0.21.0"):
        shift, _, _ = _phase_cross_correlation(img1, img2, **kw)
        return shift
    return _phase_cross_correlation(
        img1, img2, return_error=False, **kw  # deprecated in 0.21 and removed in 0.22
    )


class SkimageCorrelationIntensityMapping(
    IntensityMapping,
    registry_id=IntensityMapping.RegistryId("CrossCorrelation", "SciKitImage"),
):
    SUPPORTED_TRANSFORMATIONS = ["translation"]

    def __init__(
        self,
        transfo_type: TransformationType,
        upsample_factor: int = 5,
        normalization: bool = True,
        mask: Optional[numpy.ndarray] = None,
        **kw
    ) -> None:
        self._factor = upsample_factor
        if normalization:
            self._normalization = "phase"
        else:
            self._normalization = None
        self._mask = mask
        super().__init__(transfo_type, **kw)

    def identity(self, dimension: int = 2) -> SciKitImageHomography:
        return SciKitImageHomography(
            numpy.identity(dimension + 1), TransformationType.identity
        )

    def calculate(
        self,
        from_image: numpy.ndarray,
        to_image: numpy.ndarray,
    ) -> SciKitImageHomography:
        if self.type == self.type.translation:
            if self._mask is not None and numpy.any(self._mask):
                self._mask = self._mask.astype(bool)
                shift = phase_cross_correlation(
                    from_image,
                    to_image,
                    moving_mask=self._mask,
                    reference_mask=self._mask,
                )
            else:
                shift = phase_cross_correlation(
                    from_image,
                    to_image,
                    normalization=self._normalization,
                    upsample_factor=self._factor,
                )

            passive = numpy.identity(from_image.ndim + 1)
            passive[0:-1, -1] = shift
            return SciKitImageHomography(passive, transfo_type="translation")
        elif self.type == self.type.similarity:
            # magnitude of FFT
            from_ft = numpy.abs(fftshift(fft2(from_image)))
            to_ft = numpy.abs(fftshift(fft2(to_image)))

            # transform magnitudes into log-polar
            radius = min(from_image.shape) // 4
            warped_from_ft = warp_polar(from_ft, radius=radius, scaling="log", order=1)
            warped_to_ft = warp_polar(to_ft, radius=radius, scaling="log", order=1)

            # half the log-polar fourier magnitudes and calculate their relative shift
            warped_from_ft = warped_from_ft[: warped_from_ft.shape[0] // 2, :]
            warped_to_ft = warped_to_ft[: warped_to_ft.shape[0] // 2, :]
            shifts = phase_cross_correlation(
                warped_from_ft,
                warped_to_ft,
                normalization=self._normalization,
                upsample_factor=self._factor,
            )

            # recover scale and angle from the shift
            recovered_angle = shifts[0] / 180 * numpy.pi
            scale = numpy.exp(-shifts[1] * numpy.log(radius) / radius)

            # apply scale and angle as a centered transform
            similarity = SimilarityTransform(scale=scale, rotation=-recovered_angle)
            translation = SimilarityTransform(
                translation=(from_image.shape[0] / 2, from_image.shape[1] / 2)
            )
            mtranslation = SimilarityTransform(
                translation=(-from_image.shape[0] / 2, -from_image.shape[1] / 2)
            )
            full = mtranslation + similarity + translation
            return SciKitImageHomography(full.params, transfo_type="similarity")

        else:
            raise ValueError(
                "Only translation possible with SciKitImage Phase Correlation"
            )



import pytest
from ewoksorange.tests.conftest import qtapp  # noqa F401
from ewoksorange.canvas.handler import OrangeCanvasHandler


@pytest.fixture(scope="session")
def ewoks_orange_canvas(qtapp):  # noqa F811
    with OrangeCanvasHandler() as handler:
        yield handler



import numpy
import pytest
from ..features.mapping.lstsq import get_lstsq_solver

try:
    from ..features.mapping.scikitimage_backend import get_ransac_solver
except ImportError:
    get_ransac_solver = None
from ..transformation.numpy_backend import homography_transform_coordinates

SOLVERS = {"lstsq": get_lstsq_solver, "ransac": get_ransac_solver}
OPTIONS = {"lstsq": dict(), "ransac": {"min_samples": 8, "residual_threshold": 2}}


@pytest.fixture
def random_state(scope="module"):
    return numpy.random.RandomState(seed=100)


@pytest.mark.parametrize("solver", list(SOLVERS))
def test_solver_translation(solver, random_state):
    func = SOLVERS[solver]
    if func is None:
        pytest.skip("dependencies not installed")
    options = OPTIONS[solver]

    from_coordinates = random_state.uniform(-10, 10, (2, 10))

    active = numpy.identity(3, dtype=numpy.float32)
    active[0:2, 2] = [1, 2]

    to_coordinates = homography_transform_coordinates(active, from_coordinates)
    active2 = func("translation", **options)(from_coordinates, to_coordinates)
    numpy.testing.assert_allclose(active, active2)


@pytest.mark.parametrize("solver", list(SOLVERS))
def test_solver_rigid(solver, random_state):
    func = SOLVERS[solver]
    if func is None:
        pytest.skip("dependencies not installed")
    options = OPTIONS[solver]

    from_coordinates = random_state.uniform(-10, 10, (2, 10))

    active = numpy.identity(3, dtype=numpy.float32)
    active[0:2, 2] = [1, 2]
    active[0, 0] = numpy.cos(0.1)
    active[1, 0] = numpy.sin(0.1)
    active[0, 1] = -numpy.sin(0.1)
    active[1, 1] = numpy.cos(0.1)

    to_coordinates = homography_transform_coordinates(active, from_coordinates)
    active2 = func("rigid", **options)(from_coordinates, to_coordinates)
    numpy.testing.assert_allclose(active, active2, rtol=1e-2)


@pytest.mark.parametrize("solver", list(SOLVERS))
def test_solver_similarity(solver, random_state):
    func = SOLVERS[solver]
    if func is None:
        pytest.skip("dependencies not installed")
    options = OPTIONS[solver]

    from_coordinates = random_state.uniform(-10, 10, (2, 10))

    active = numpy.identity(3, dtype=numpy.float32)
    active[0:2, 2] = [1, 2]
    active[0, 0] = 1.1 * numpy.cos(0.1)
    active[1, 0] = numpy.sin(0.1)
    active[0, 1] = -numpy.sin(0.1)
    active[1, 1] = 1.1 * numpy.cos(0.1)

    to_coordinates = homography_transform_coordinates(active, from_coordinates)
    active2 = func("similarity", **options)(from_coordinates, to_coordinates)
    numpy.testing.assert_allclose(active, active2)


@pytest.mark.parametrize("solver", list(SOLVERS))
def test_solver_affine(solver, random_state):
    func = SOLVERS[solver]
    if func is None:
        pytest.skip("dependencies not installed")
    options = OPTIONS[solver]

    from_coordinates = random_state.uniform(-10, 10, (2, 10))

    active = numpy.identity(3, dtype=numpy.float32)
    active[0:2, 2] = [1, 2]
    active[0, 0] = 1.1 * numpy.cos(0.1)
    active[1, 0] = numpy.sin(0.1)
    active[0, 1] = -numpy.sin(0.1)
    active[1, 1] = 1.2 * numpy.cos(0.1)

    to_coordinates = homography_transform_coordinates(active, from_coordinates)
    active2 = func("affine", **options)(from_coordinates, to_coordinates)
    numpy.testing.assert_allclose(active, active2)


@pytest.mark.parametrize("solver", list(SOLVERS))
def test_solver_projective(solver, random_state):
    func = SOLVERS[solver]
    if func is None:
        pytest.skip("dependencies not installed")
    options = OPTIONS[solver]

    from_coordinates = random_state.uniform(-10, 10, (2, 10))

    active = numpy.identity(3, dtype=numpy.float32)
    active[2, 0] = 0.2

    to_coordinates = homography_transform_coordinates(active, from_coordinates)
    active2 = func("projective", **options)(from_coordinates, to_coordinates)
    numpy.testing.assert_allclose(active, active2, atol=1e-10)



import numpy
from ewoksorange.tests.utils import execute_task

from orangecontrib.ewoksndreg.reg2d_features import OWReg2DFeatures
from orangecontrib.ewoksndreg.reg2d_intensities import OWReg2DIntensities
from orangecontrib.ewoksndreg.reg2d_transform import OWReg2DTransform
from orangecontrib.ewoksndreg.example2d_stack import OWExample2DStack
from orangecontrib.ewoksndreg.reg2d_preeval import OWReg2DPreEvaluation
from .data import data_for_registration
from ..io import output_stack
from ..transformation.numpy_backend import NumpyHomography


def test_owreg2d_features_task(tmpdir):
    _test_owreg2d_features(tmpdir, OWReg2DFeatures.ewokstaskclass)


def test_owreg2d_features_widget(tmpdir, qtapp):
    _test_owreg2d_features(tmpdir, OWReg2DFeatures)


def _test_owreg2d_features(tmpdir, task_cls):
    file_path = str(tmpdir / "data.h5")
    data_path = "entry/instrument/detector/data"
    url = f"silx://{file_path}::{data_path}"
    istack, active1, passive1 = data_for_registration.images("translation", plot=0)
    with output_stack.output_context(url) as stacko:
        stacko.add_points(istack)

    result = execute_task(
        task_cls,
        inputs={
            "imagestack": url,
            "transformation_type": "translation",
            "detector": ("Sift", "SciKitImage"),
            "matcher": ("Descriptor", "SciKitImage"),
            "mapper": ("LstSq", "Numpy"),
        },
    )

    active2 = numpy.stack([tr.active for tr in result["transformations"]])
    passive2 = numpy.stack([tr.passive for tr in result["transformations"]])

    numpy.testing.assert_allclose(active1, active2, rtol=0.01)
    numpy.testing.assert_allclose(passive1, passive2, rtol=0.01)


def test_owreg2d_intensities_task(tmpdir):
    _test_owreg2d_intensities(tmpdir, OWReg2DIntensities.ewokstaskclass)


def test_owreg2d_intensities_widget(tmpdir, qtapp):
    _test_owreg2d_intensities(tmpdir, OWReg2DIntensities)


def _test_owreg2d_intensities(tmpdir, task_cls):
    file_path = str(tmpdir / "data.h5")
    data_path = "entry/instrument/detector/data"
    url = f"silx://{file_path}::{data_path}"
    istack, active1, passive1 = data_for_registration.images("translation", plot=0)
    with output_stack.output_context(url) as stacko:
        stacko.add_points(istack)

    result = execute_task(
        task_cls,
        inputs={
            "imagestack": url,
            "transformation_type": "translation",
            "mapper": ("CrossCorrelation", "Numpy"),
        },
    )

    active2 = numpy.stack([tr.active for tr in result["transformations"]])
    passive2 = numpy.stack([tr.passive for tr in result["transformations"]])

    numpy.testing.assert_allclose(active1, active2, rtol=0.01)
    numpy.testing.assert_allclose(passive1, passive2, rtol=0.01)


def test_owreg2d_intensities_multi_task(tmpdir):
    _test_owreg2d_intensities_multi(tmpdir, OWReg2DIntensities.ewokstaskclass)


def test_owreg2d_intensities_multi_widget(tmpdir, qtapp):
    _test_owreg2d_intensities_multi(tmpdir, OWReg2DIntensities)


def _test_owreg2d_intensities_multi(tmpdir, task_cls):
    file_path = str(tmpdir / "data.h5")
    data_path = "entry/instrument/detector/data"
    url = f"silx://{file_path}::{data_path}"
    istack, active1, passive1 = data_for_registration.images("translation", plot=0)
    istack = [numpy.asarray(istack)] * 3
    with output_stack.output_context(url) as stacko:
        stacko.add_points(istack)

    result = execute_task(
        task_cls,
        inputs={
            "imagestack": url,
            "transformation_type": "translation",
            "mapper": ("CrossCorrelation", "Numpy"),
        },
    )
    for i in range(3):
        active2 = numpy.stack([tr.active for tr in result["transformations"][i]])
        passive2 = numpy.stack([tr.passive for tr in result["transformations"][i]])

        numpy.testing.assert_allclose(active1, active2, rtol=0.01)
        numpy.testing.assert_allclose(passive1, passive2, rtol=0.01)


def test_owreg2d_apply_task():
    _test_owreg2d_apply(OWReg2DTransform.ewokstaskclass)


def test_owreg2d_apply_widget(qtapp):
    _test_owreg2d_apply(OWReg2DTransform)


def _test_owreg2d_apply(task_cls):
    istack, active, passive = data_for_registration.images("translation", plot=0)

    forward = [NumpyHomography(M) for M in passive]
    backward = [NumpyHomography(M) for M in active]

    result = execute_task(
        task_cls,
        inputs={"imagestack": istack, "transformations": forward},
    )
    result = execute_task(
        task_cls,
        inputs={"imagestack": result["imagestack"], "transformations": backward},
    )

    istack = numpy.asarray(istack)
    ostack = numpy.asarray(result["imagestack"])
    ostack[0:9, :] = istack[0:9, :]
    ostack[:, 0:6] = istack[:, 0:6]

    numpy.testing.assert_allclose(istack, ostack)


def test_owreg2d_apply_multi_task():
    _test_owreg2d_apply_multi(OWReg2DTransform.ewokstaskclass)


def test_owreg2d_apply_multi_widget(qtapp):
    _test_owreg2d_apply_multi(OWReg2DTransform)


def _test_owreg2d_apply_multi(task_cls):
    istack, active, passive = data_for_registration.images("translation", plot=0)
    istack = numpy.tile(istack, (4, 1, 1, 1))

    forward = [[NumpyHomography(M) for M in passive]]
    backward = [[NumpyHomography(M) for M in active]]

    result = execute_task(
        task_cls,
        inputs={"imagestack": istack, "transformations": forward},
    )
    result = execute_task(
        task_cls,
        inputs={"imagestack": result["imagestack"], "transformations": backward},
    )

    istack = numpy.asarray(istack)
    ostack = numpy.asarray(result["imagestack"])
    ostack[:, 0:9, :] = istack[:, 0:9, :]
    ostack[:, :, 0:6] = istack[:, :, 0:6]

    numpy.testing.assert_allclose(istack, ostack)


def test_owexample2d_stack_task():
    _test_owexample2d_stack(OWExample2DStack.ewokstaskclass)


def test_owexample2d_stack_widget(qtapp):
    _test_owexample2d_stack(OWExample2DStack.ewokstaskclass)


def _test_owexample2d_stack(task_cls):
    istack, active, passive = data_for_registration.images("translation", plot=0)

    result = execute_task(
        task_cls,
        inputs={"name": "astronaut", "transformation_type": "translation"},
    )

    istack = numpy.asarray(istack)
    ostack = numpy.asarray(result["imagestack"])
    numpy.testing.assert_allclose(istack, ostack)

    for tr, a, p in zip(result["transformations"], active, passive):
        numpy.testing.assert_allclose(tr.active, a)
        numpy.testing.assert_allclose(tr.passive, p)


def test_owreg2d_preeval_widget(tmpdir):
    _test_owreg2d_preeval(tmpdir, OWReg2DPreEvaluation)


def test_owreg2d_preeval_task(tmpdir):
    _test_owreg2d_preeval(tmpdir, OWReg2DPreEvaluation.ewokstaskclass)


def _test_owreg2d_preeval(tmpdir, task_cls):
    file_path = str(tmpdir / "data.h5")
    data_path = "entry/instrument/detector/data"
    url = f"silx://{file_path}::{data_path}"
    istack = data_for_registration.noisy_imagestacks("rigid", nstacks=4)

    with output_stack.output_context(url) as stacko:
        stacko.add_points(istack)

    result = execute_task(
        task_cls,
        inputs={
            "imagestacks": url,
        },
    )
    numpy.testing.assert_equal(result["ranking"], [0, 1, 2, 3])



from typing import Optional
import numpy
from ...registry import Registered
from ..features import Features

__all__ = ["FeatureDetector"]


class FeatureDetector(Registered, register=False):
    def __init__(self, mask: Optional[numpy.ndarray] = None) -> None:
        self._mask = mask

    def find(self, arr: numpy.ndarray) -> Features:
        raise NotImplementedError



from contextlib import contextmanager

from silx.gui.plot import Plot2D
from silx.gui.widgets.FrameBrowser import HorizontalSliderWithBrowser

from ewoksorange.bindings import OWEwoksWidgetOneThread
from ewoksorange.bindings import ow_build_opts
from ewoksorange.gui.parameterform import ParameterForm

from ewoksndreg.tasks import Example2DStack
from ewoksndreg.transformation.types import TransformationType
from ewoksndreg.io.input_stack import input_context

try:
    import skimage
except ImportError:
    skimage = None


__all__ = ["OWExample2DStack"]


class OWExample2DStack(
    OWEwoksWidgetOneThread, **ow_build_opts, ewokstaskclass=Example2DStack
):
    name = "2D Example Stack"
    description = "Generate a stack of 2D example images"
    icon = "icons/load_stack.svg"

    def __init__(self):
        super().__init__()
        self._init_control_area()
        self._init_main_area()

    def _init_control_area(self):
        super()._init_control_area()
        self._default_inputs_form = ParameterForm(parent=self.controlArea)
        values = self.get_default_input_values(include_missing=True)

        options = {
            "name": {
                "value_for_type": [
                    "astronaut",
                    "camera",
                    "brick",
                    "grass",
                    "gravel",
                    "cell",
                ]
            },
            "transformation_type": {
                "value_for_type": list(TransformationType.__members__),
                "serialize": str,
            },
            "nimages": {"value_for_type": 0},
            "shape": {
                "value_for_type": "",
                "serialize": lambda tpl: ",".join(list(map(str, tpl))),
                "deserialize": lambda s: tuple(map(int, s.split(","))),
            },
            "add_noise_dim": {
                "value_for_type": False,
                "enabled": bool(skimage),
                "label": "Add noise dimension",
            },
        }

        for name, kw in options.items():
            self._default_inputs_form.addParameter(
                name,
                value=values[name],
                value_change_callback=self._default_inputs_changed,
                **kw,
            )

    def _default_inputs_changed(self):
        self.update_default_inputs(**self._default_inputs_form.get_parameter_values())
        self._update_input_data()

    def handleNewSignals(self) -> None:
        self._update_input_data()
        super().handleNewSignals()

    def task_output_changed(self):
        self._update_output_data()

    def _init_main_area(self):
        super()._init_main_area()
        layout = self._get_main_layout()
        self._plot = Plot2D(parent=self.mainArea)
        layout.addWidget(self._plot)
        self._slider1 = HorizontalSliderWithBrowser(parent=self.mainArea)
        layout.addWidget(self._slider1)
        self._slider1.valueChanged[int].connect(lambda: self._select_output_image())
        self._slider2 = HorizontalSliderWithBrowser(parent=self.mainArea)
        layout.addWidget(self._slider2)
        self._slider2.valueChanged[int].connect(lambda: self._select_output_image())
        self._slider2.setVisible(False)
        self._update_output_data()

    def _update_input_data(self):
        dynamic = self.get_dynamic_input_names()
        for name in self.get_input_names():
            self._default_inputs_form.set_parameter_enabled(name, name not in dynamic)

    @contextmanager
    def _output_context(self, images=None):
        if images is not None:
            yield images
            return
        try:
            with input_context(self.get_task_output_value("imagestack")) as images:
                yield images
        except TypeError:
            yield None

    def _update_output_data(self):
        with self._output_context() as images:
            if images is None:
                return

            assert images.ndim in (3, 4)

            if images.ndim == 4:
                self._slider1.setMaximum(max(images.shape[0] - 1, 0))
                self._slider2.setVisible(True)
                self._slider2.setMaximum(max(images.shape[1] - 1, 0))
            else:
                self._slider2.setVisible(False)
                self._slider1.setMaximum(max(images.shape[0] - 1, 0))
            self._select_output_image(images=images)

    def _select_output_image(self, images=None):
        if self._slider2.isVisible():
            selection = (self._slider1.value(), self._slider2.value())
        else:
            selection = self._slider1.value()
        with self._output_context(images=images) as images:
            if images:
                self._plot.addImage(images[selection], legend="image")



import numpy
from ..transformation import apply_transformations
from ..transformation.homography import Homography
from ..io.input_stack import InputStackNumpy
from ..io.output_stack import OutputStackNumpy
from .data import data_for_registration
import pytest

_HOMOGRAPHIES = {
    f"homography{'_'.join(k)}": v for k, v in Homography.get_subclass_items()
}


@pytest.mark.parametrize("homography", _HOMOGRAPHIES)
def test_apply_transformations(homography):
    istack, active, passive = data_for_registration.images("translation", plot=0)

    forward = [_HOMOGRAPHIES[homography](passive=M) for M in passive]
    backward = [_HOMOGRAPHIES[homography](passive=M) for M in active]

    data1 = list()
    with OutputStackNumpy(data1) as ostack:
        apply_transformations(istack, ostack, forward, interpolation_order=0)

    data2 = list()
    with OutputStackNumpy(data2) as ostack:
        with InputStackNumpy(data1) as istack2:
            apply_transformations(istack2, ostack, backward, interpolation_order=0)

    istack = numpy.asarray(istack)
    ostack = numpy.asarray(data2)
    idx = numpy.isnan(ostack)
    ostack[idx] = istack[idx]

    numpy.testing.assert_allclose(istack, ostack)


@pytest.mark.parametrize("homography", _HOMOGRAPHIES)
def test_apply_translation(homography):
    image = numpy.random.uniform(0.0, 10, (10, 10))

    transform = numpy.array([[1, 0, 6], [0, 1, -2], [0, 0, 1]], dtype=numpy.float64)
    forward = _HOMOGRAPHIES[homography](passive=transform)
    goal = numpy.zeros((10, 10))
    goal[0:4, 2:10] = image[6:10, 0:8]
    actual = forward.apply_data(image, cval=0)

    numpy.testing.assert_allclose(actual, goal)


@pytest.mark.parametrize("homography", _HOMOGRAPHIES)
def test_apply_coordinate(homography):
    nx, ny = numpy.meshgrid(numpy.arange(4, dtype=numpy.float64), numpy.arange(4))
    nx, ny = nx.flatten(), ny.flatten()
    matrix = numpy.identity(3)
    matrix[0:2, 2] = [-1, -3]

    forward = _HOMOGRAPHIES[homography](passive=matrix)
    res1 = forward.apply_coordinates(numpy.asarray([nx, ny]))
    numpy.testing.assert_allclose([nx + 1, ny + 3], res1)

    matrix = numpy.array([[0, -1, 0], [1, 0, 0], [0, 0, 1]], dtype=numpy.float64)

    forward = _HOMOGRAPHIES[homography](passive=matrix)
    res1 = forward.apply_coordinates(numpy.asarray([nx, ny]))

    numpy.testing.assert_allclose([ny, -nx], res1, atol=numpy.finfo(numpy.float64).eps)



import numpy as np
from typing import Tuple, Union


def resize(image: np.ndarray, out_shape: Tuple[int], order: int = 0) -> np.ndarray:
    if image.ndim != len(out_shape):
        raise ValueError("New shape must have same dimensionality as image")

    in_shape = image.shape

    ranges = [np.arange(0, shp, shp / out_shape[i]) for i, shp in enumerate(in_shape)]
    X = np.meshgrid(*ranges[::-1])
    X = np.array(X).astype(int)
    X = X[::-1]
    if order == 0:
        result = image[X[0], X[1]]
        result = np.reshape(result, out_shape)
        return result


def rescale(
    image: np.ndarray, factor: Union[float, int, Tuple], order: int = 0
) -> np.ndarray:
    shp = image.shape

    factor = float(factor) if isinstance(factor, int) else factor

    if isinstance(factor, float):
        factor = (factor,) * image.ndim

    if len(factor) != image.ndim:
        raise ValueError(
            f"{len(factor)} scaling factors must match image dimension {image.ndim}"
        )

    new_shape = tuple([int(x * factor[i]) for i, x in enumerate(shp)])

    return resize(image, new_shape, order=order)



"""Homography abstraction.

A homography transformation can be represented be the following change-of-frame matrix

    $$
    F=\\begin{bmatrix}
    C_{n\\times n}&T_{n\\times 1}\\\\
    P_{1\\times n}&1
    \\end{bmatrix}
    $$

Different transformation types can be distinguished based on the properties of the translation vector $T$, change-of-basis matrix $C$ and projection vector $P$

    * translation:
        * $C$ is the identity matrix and $P$ are all zeros
    * proper rigid (euclidian) transformation:
        * $P$ are all zeros
        * $C$ is any orthogonal matrix ($C^T=C^{-1}$) with determinant $\\det{C}=+1$
        * preserves angles, distances and handedness
        * translation + rotation
    * rigid (euclidian) transformation:
        * $P$ are all zeros
        * $C$ is any orthogonal matrix ($C^T=C^{-1}$)
        * preserves angles and distances
        * proper rigid transformation + reflection
    * similarity transformation:
        * $P$ are all zeros
        * $C=rA$ where A any orthogonal matrix ($A^T=A^{-1}$) and $r>0$
        * preserves angles and ratios between distances
        * rigid transformation + isotropic scaling
    * affine transformation:
        * $P$ are all zeros
        * $C$ is any invertible matrix (i.e. linear transformation)
        * preserves parallelism
        * similarity transformation + non-isotropic scaling and shear
    * projective transformation (homography):
        * $C$ is any invertible matrix (i.e. linear transformation)
        * preserves collinearity
        * affine transformation + projection
"""

from typing import Optional
import numpy

from .types import TransformationType
from .base import Transformation


def type_from_matrix(matrix: numpy.ndarray) -> TransformationType:
    """
    :param active: shape `(N+1, N+1)` or `(K, N+1, N+1)`
    :returns: transformation type
    """
    N = matrix.shape[0] - 1
    if matrix.shape != (N + 1, N + 1):
        raise ValueError("must be a square matrix")
    if numpy.allclose(matrix, numpy.identity(N + 1)):
        return TransformationType.identity
    if matrix[-1, -1] != 1 or not numpy.allclose(matrix[-1, 0:-1], 0):
        return TransformationType.projective
    C = matrix[:-1, :-1]
    if numpy.allclose(C, numpy.identity(N)):
        if numpy.allclose(matrix[-1, :-1], 0):
            return TransformationType.translation
        else:
            return TransformationType.rigid
    C_T = C.T
    C_I = numpy.linalg.inv(C)
    if numpy.allclose(numpy.linalg.det(C), 1, rtol=1e-7):
        return TransformationType.rigid
    C_I[C_I == 0] = 1
    r = C_T / C_I
    if numpy.allclose(r, r.mean()):
        return TransformationType.similarity
    return TransformationType.affine


class Homography(Transformation, register=False):
    def __init__(
        self,
        passive: numpy.ndarray,
        transfo_type: Optional[TransformationType] = None,
    ) -> None:
        if transfo_type is None:
            transfo_type = type_from_matrix(passive)
        self._passive = passive
        self._active = None
        super().__init__(transfo_type)

    @property
    def passive(self) -> numpy.ndarray:
        return self._passive

    @property
    def active(self) -> numpy.ndarray:
        if self._active is None:
            self._active = numpy.linalg.inv(self._passive)
        return self._active

    def as_parameters(self) -> numpy.ndarray:
        return params_from_trans(self)

    def __mul__(self, other: Transformation) -> Transformation:
        if isinstance(other, Homography):
            if self.passive.shape == other.passive.shape:
                return Homography(self.passive @ other.passive)
            else:
                raise TypeError("Homographies must have same dimensions")
        else:
            raise TypeError(
                "Concatenating Homography and non-homography is not possible"
            )


"""
Conversions between matrix representation of transformation and tuple of parameters of transformation
Identity: ()
translation: (shift_x, shift_y)
proper_rigid: (angle, shift_x, shift_y)
affine: (matrix, translation)
projective: flattened version of matrix
"""


def params_from_trans(transformation: Homography) -> list:
    if transformation._type == "identity":
        return []

    elif transformation._type == "translation":
        return list(transformation.passive[1::-1, 2])

    elif transformation._type == "rigid":
        return [
            numpy.arccos(transformation.passive[0, 0]),
            transformation.passive[1, 2],
            transformation.passive[0, 2],
        ]

    elif transformation._type == "affine" or transformation._type == "similarity":
        return list(transformation.passive[0:2].flatten())

    elif transformation._type == "projective":
        return list(transformation.passive.flatten())

    else:
        raise NotImplementedError


def matrix_from_params(
    params: list, transformation_type: TransformationType
) -> numpy.ndarray:
    if transformation_type == "identity":
        return numpy.identity(3)

    elif transformation_type == "translation":
        matrix = numpy.identity(3)
        matrix[0:2, 2] = params[::-1]
        return matrix

    elif transformation_type == "rigid":
        cos, sin = numpy.cos(params[0]), numpy.sin(params[0])
        matrix = numpy.asarray(
            [[cos, -sin, params[2]], [sin, cos, params[1]], [0, 0, 1]]
        )
        return matrix

    elif transformation_type == "affine":
        matrix = numpy.identity(3)
        matrix[0:2] = numpy.reshape(params, (2, 3))
        return matrix

    elif transformation_type == "projective":
        return numpy.reshape(params, (3, 3))

    else:
        raise NotImplementedError


def reverse_indices(matrix: numpy.ndarray):
    N = matrix.shape[0]
    if matrix.ndim != 2 and matrix.shape != (N, N):
        raise ValueError("Input must be square matrix")
    switched = matrix.copy()
    switched[0:-1, 0:-1] = switched[0:-1, 0:-1].T
    switched[:-1, -1] = switched[:-1, -1][::-1]
    switched[-1, :-1] = switched[-1, :-1][::-1]
    return switched



"""Data features in n-D datasets"""

import numpy
from .base import Features
from .numpy_backend import NumpyKeypointFeatures

try:
    from .silx_backend import SilxDescriptorFeatures  # noqa F401
except ImportError:
    pass

try:
    from .scikitimage_backend import SciKitImageDescriptorFeatures  # noqa F401
except ImportError:
    pass


def _to_numpy_keypoint_features(features: Features) -> NumpyKeypointFeatures:
    return NumpyKeypointFeatures(numpy.asarray(features.coordinates).T)


for cls in Features.get_subclasses():
    if cls is not NumpyKeypointFeatures:
        cls._set_feature_type_converter(
            NumpyKeypointFeatures, _to_numpy_keypoint_features
        )


try:

    def _scikitimage_to_silx_descriptor(
        skimage_features: SciKitImageDescriptorFeatures,
    ) -> SilxDescriptorFeatures:
        desc = skimage_features.skimage_descriptors
        y, x = skimage_features.coordinates

        desc_shape = desc.shape
        if desc_shape[1] != 128:
            desc = desc.reshape((desc_shape[0], 128, desc_shape[1] // 128))
            desc = desc.astype(numpy.uint8)
            desc = desc.sum(axis=-1)

        dtype = numpy.dtype(
            [
                ("x", numpy.float32),
                ("y", numpy.float32),
                ("scale", numpy.float32),
                ("angle", numpy.float32),
                ("desc", (numpy.uint8, 128)),
            ]
        )

        n = len(desc)
        silx_features = numpy.recarray(shape=(n,), dtype=dtype)
        zeros = numpy.zeros(n, numpy.float32)
        silx_features[:].x = x.astype(numpy.float32)
        silx_features[:].y = y.astype(numpy.float32)
        silx_features[:].scale = zeros
        silx_features[:].angle = zeros
        silx_features[:].desc = desc.astype(numpy.uint8)
        return SilxDescriptorFeatures(silx_features)

    def _silx_to_scikitimage_descriptor(
        silx_features: SilxDescriptorFeatures,
    ) -> SciKitImageDescriptorFeatures:
        keypoints = silx_features.coordinates.T
        descriptors = silx_features.silx_features.desc
        return SciKitImageDescriptorFeatures(keypoints, descriptors)

    SciKitImageDescriptorFeatures._set_feature_type_converter(
        SilxDescriptorFeatures, _scikitimage_to_silx_descriptor
    )
    SilxDescriptorFeatures._set_feature_type_converter(
        SciKitImageDescriptorFeatures, _silx_to_scikitimage_descriptor
    )

except NameError:
    pass



from ewokscore.task import Task
from ..io.input_stack import input_context
from ..io.output_stack import output_context

__all__ = ["SelectSlice"]


class SelectSlice(
    Task,
    input_names=["imagestack"],
    optional_input_names=["inputs_are_stacks", "url", "x", "y", "z"],
    output_names=["imagestack"],
):
    """Slice out a subvolume of the imagestack"""

    def run(self):
        url = self.get_input_value("url", None)
        with output_context(url=url) as ostack:
            with input_context(
                self.inputs.imagestack,
                inputs_are_stacks=self.get_input_value("inputs_are_stacks", None),
            ) as istack:
                shape = istack.shape

                x = self.get_input_value("x", f"0:{shape[1]}")
                y = self.get_input_value("y", f"0:{shape[2]}")
                z = self.get_input_value("z", f"0:{shape[0]}")

                x = parse(x)
                y = parse(y)
                z = parse(z)
                ostack.add_points([img[y, x] for img in istack[z]])

                if url:
                    self.outputs.imagestack = url
                else:
                    self.outputs.imagestack = ostack.data


def parse(string: str):
    return slice(
        *map(lambda x: int(x.strip()) if x.strip() else None, string.split(":"))
    )



"""Determine mapping betweem two sets of features"""

from .base import *  # noqa F401
from .numpy_backend import *  # noqa F401

try:
    from .scipy_backend import *  # noqa F401
except ImportError:
    pass

try:
    from .scikitimage_backend import *  # noqa F401
except ImportError:
    pass



from typing import Optional, Sequence, Dict
import numpy

from ewoksndreg.transformation.homography import type_from_matrix, reverse_indices
from ewoksndreg.transformation.base import Transformation
from ewoksndreg.transformation.types import TransformationType
import SimpleITK as sitk

__all__ = ["SimpleITKTransformation"]


class SimpleITKTransformation(
    Transformation, registry_id=Transformation.RegistryId("Transformation", "SimpleITK")
):
    def __init__(
        self,
        transfo_type: Optional[TransformationType] = None,
        warp_options: Optional[Dict] = None,
        transformation: Optional[sitk.Transform] = None,
        passive: Optional[numpy.ndarray] = None,
        displacement_field: Optional[numpy.ndarray] = None,
        bspline: Optional[numpy.ndarray] = None,
    ) -> None:
        if warp_options is None:
            warp_options = dict()
        self._warp_options = warp_options
        self._passive = None
        if transformation:
            self._sitk_passive = transformation
            super().__init__(self.type_from_transform(transformation))

        elif displacement_field is not None:
            super().__init__("displacement_field")
            self._sitk_passive = sitk.DisplacementFieldTransform(
                sitk.GetImageFromArray(displacement_field, True)
            )

        elif bspline is not None:
            super().__init__("bspline")
            spline_order = bspline[-1]
            self._sitk_passive = sitk.BSplineTransform(bspline[:-1], spline_order)

        elif passive is not None:
            if transfo_type is None:
                transfo_type = type_from_matrix(passive)
            super().__init__(transfo_type)
            matrix = passive[:-1]

            dim = matrix.shape[0]
            if self.type == self.type.identity:
                self._sitk_passive = sitk.TranslationTransform(dim)
            elif self.type == self.type.translation:
                self._sitk_passive = sitk.TranslationTransform(dim)
                self._sitk_passive.SetOffset(matrix[:, -1][::-1])

            elif self.is_homography():
                if self.type == self.type.rigid and dim == 2:
                    self._sitk_passive = sitk.Euler2DTransform()
                elif self.type == self.type.rigid and dim == 3:
                    self._sitk_passive = sitk.Euler3DTransform()
                elif dim == 2 and (
                    self.type == self.type.similarity or self._type == self._type.rigid
                ):
                    self._sitk_passive = sitk.Similarity2DTransform()
                elif dim == 3 and (
                    self.type == self.type.similarity or self._type == self._type.rigid
                ):
                    self._sitk_passive = sitk.Similarity3DTransform()
                elif self.type == self.type.affine:
                    self._sitk_passive = sitk.AffineTransform(dim)
                self._sitk_passive.SetCenter(dim * [0])
                self._sitk_passive.SetMatrix(tuple(matrix[:, 0:-1].T.flatten()))
                self._sitk_passive.SetTranslation(tuple(matrix[:, -1][::-1]))

        try:
            self._sitk_passive
        except AttributeError:
            self._sitk_passive = self.get_identity(self.type)

    @property
    def displacement_field(self) -> numpy.ndarray:
        try:
            return sitk.GetArrayFromImage(self._sitk_passive.GetDisplacementField())
        except AttributeError:
            raise AttributeError(
                "Tried to get displacement field but transformation type is {self.type}"
            )

    @property
    def passive(self) -> numpy.ndarray:
        if self._passive is not None:
            return self._passive
        elif self.is_homography():
            return self._add_passive()
        else:
            raise AttributeError(
                "the passive matrix is only available if the transformation type can be represented by a 2x3 matrix"
            )

    def get_identity(self, transfo_type: TransformationType, dim: int = 2):
        if transfo_type == "translation":
            transform = sitk.TranslationTransform(dim)
        if transfo_type == "rigid":
            transform = sitk.Euler2DTransform() if dim == 2 else sitk.Euler3DTransform()
        if transfo_type == "similarity":
            transform = (
                sitk.Similarity2DTransform()
                if dim == 2
                else sitk.Similarity3DTransform()
            )
        if transfo_type == "affine":
            transform = sitk.AffineTransform(dim)
        if transfo_type in ["bspline", "displacement_field"]:
            raise ValueError(
                "Can't create empty displacement field without image specifications"
            )
        return transform

    def type_from_transform(self, transformation: sitk.Transform):
        if isinstance(transformation, sitk.TranslationTransform):
            transfo_type = "translation"
        elif isinstance(transformation, (sitk.Euler2DTransform, sitk.Euler3DTransform)):
            transfo_type = "rigid"
        elif isinstance(
            transformation, (sitk.Similarity2DTransform, sitk.Similarity3DTransform)
        ):
            transfo_type = "similarity"
        elif isinstance(transformation, (sitk.AffineTransform)):
            transfo_type = "affine"
        elif isinstance(transformation, sitk.DisplacementFieldTransform):
            transfo_type = "displacement_field"
        elif isinstance(transformation, sitk.BSplineTransform):
            transfo_type = "bspline"
        else:
            transfo_type = "composite"
        return transfo_type

    def _add_passive(self):
        passive = numpy.identity(3)
        if self._type == self._type.identity:
            self._passive = passive

        elif self._type == self._type.translation:
            passive[1::-1, 2] = self._sitk_passive.GetOffset()
            self._passive = passive

        elif self.is_homography():
            cx, cy = self._sitk_passive.GetCenter()
            A = self._sitk_passive.GetMatrix()
            tx, ty = self._sitk_passive.GetTranslation()
            matrix = numpy.asarray([[A[0], A[1], 0], [A[2], A[3], 0], [0, 0, 1]])
            pre = numpy.asarray([[1, 0, -cx], [0, 1, -cy], [0, 0, 1]])
            post = numpy.asarray([[1, 0, tx + cx], [0, 1, ty + cy], [0, 0, 1]])
            passive = post @ matrix @ pre
            self._passive = reverse_indices(passive)
        return self._passive

    def apply_coordinates(self, coord: Sequence[numpy.ndarray]) -> numpy.ndarray:
        """
        :param coord: shape `(N, M)`
        :returns: shape `(N, M)`
        """
        coord = coord[::-1]
        transformed = [
            self._sitk_passive.GetInverse().TransformPoint(point.astype(numpy.float64))
            for point in numpy.transpose(coord)
        ]
        return numpy.transpose(transformed)[::-1]

    def apply_data(
        self,
        data: numpy.ndarray,
        offset: Optional[numpy.ndarray] = None,
        shape: Optional[numpy.ndarray] = None,
        cval=numpy.nan,
        interpolation_order: int = 1,
    ) -> numpy.ndarray:
        """
        :param data: shape `(N1, N2, ..., M1, M2, ...)` with `len((N1, N2, ...)) = N`
        :param offset: shape `(N,)`
        :param shape: shape `(N,) = [N1', N2', ...]`
        :param cval: missing value
        :param interpolation_order: order of interpolation: 0 is nearest neighbor, 1 is bilinear,...
        :returns: shape `(N1', N2', ..., M1, M2, ...)`
        """
        kw = dict(self._warp_options)
        if shape is not None:
            kw["output_shape"] = shape
        if offset is not None:
            kw["offset"] = offset
        if cval is not None:
            kw["cval"] = cval

        if interpolation_order is not None:
            order = interpolation_order
        else:
            order = 1

        image = sitk.GetImageFromArray(data)
        resampler = sitk.ResampleImageFilter()
        resampler.SetReferenceImage(image)

        if order == 0:
            resampler.SetInterpolator(sitk.sitkNearestNeighbor)
        elif order == 1:
            resampler.SetInterpolator(sitk.sitkLinear)
        elif order == 2:
            resampler.SetInterpolator(sitk.sitkBSpline1)
        elif order == 3:
            resampler.SetInterpolator(sitk.sitkBSpline2)
        elif order == 4:
            resampler.SetInterpolator(sitk.sitkBSpline3)
        else:
            raise ValueError("Only interpolation up to order 4 possible")

        resampler.SetDefaultPixelValue(kw["cval"])
        resampler.SetTransform(self._sitk_passive)

        out = resampler.Execute(image)
        return sitk.GetArrayFromImage(out)

    def __mul__(self, other: Transformation):
        if isinstance(other, SimpleITKTransformation):
            dim = self._sitk_passive.GetDimension()
            if other._sitk_passive.GetDimension() != dim:
                raise TypeError("Transformations must have same dimensions")
            if dim == 2 and self.is_homography() and other.is_homography():
                return SimpleITKTransformation(passive=other.passive @ self.passive)
            comp = sitk.CompositeTransform(dim)
            comp.AddTransform(self._sitk_passive)
            comp.AddTransform(other._sitk_passive)
            comp.FlattenTransform()
            return SimpleITKTransformation(
                transfo_type="composite",
                warp_options=self._warp_options,
                transformation=comp,
            )
        else:
            raise TypeError(
                "SimpleITK Transformation can only be concatenated with other SimpleITK Transformation"
            )



"""Calculate active transformation between two sets of coordinates"""

from typing import Sequence
import numpy

try:
    from scipy.stats import trim_mean
except ImportError:

    def trim_mean(arr, proportiontocut, axis=0):
        arr = numpy.asarray(arr)

        if arr.size == 0:
            return numpy.nan

        if axis is None:
            arr = arr.ravel()
            axis = 0

        nobs = arr.shape[axis]
        lowercut = int(proportiontocut * nobs)
        uppercut = nobs - lowercut
        if lowercut > uppercut:
            raise ValueError("Proportion too big.")

        atmp = numpy.partition(arr, (lowercut, uppercut - 1), axis)

        sl = [slice(None)] * atmp.ndim
        sl[axis] = slice(lowercut, uppercut)
        return numpy.mean(atmp[tuple(sl)], axis=axis)


def calc_identity(
    from_coord: Sequence[numpy.ndarray], to_coord: Sequence[numpy.ndarray]
) -> numpy.ndarray:
    """Coordinate array shape is (ndim, ncoord)"""
    return numpy.identity(3, dtype=from_coord[0].dtype)


def calc_translation(
    from_coord: Sequence[numpy.ndarray], to_coord: Sequence[numpy.ndarray]
) -> numpy.ndarray:
    """Coordinate array shape is (ndim, ncoord)"""
    from_coord = numpy.asarray(from_coord)
    to_coord = numpy.asarray(to_coord)
    T = centroid(to_coord - from_coord, axis=-1)
    active = numpy.identity(3, dtype=T.dtype)
    active[0:2, 2] = T
    return active


def calc_rigid(
    from_coord: Sequence[numpy.ndarray], to_coord: Sequence[numpy.ndarray]
) -> numpy.ndarray:
    r"""Coordinate array shape is (ndim, ncoord).

    Find the proper-rigid transformation between two sets of coordinates by solving
    this system of equations (https://igl.ethz.ch/projects/ARAP/svd_rot.pdf):

    .. math::

        \begin{align*}
            X_c &= (X-Cen(X)) \\
            Y_c &= (Y-Cen(Y)) \\
            C.X + T &= X' \\
            T &= Cen(X') - C.Cen(X)
        \end{align*}

        \begin{align*}
            C.X_c &= Y_c \\
            H &= X_c . Y_c^T \\
            H &= U . S . V^T \\
            C &= V.U^T \\
        \end{align*}

        \begin{align*}
            C.X + T &= X' \\
            T &= Cen(X') - C.Cen(X)
        \end{align*}

    where `C` is found with singular value decomposition. The resulting transformation matrix is

    .. math::

        \begin{equation*}
            \begin{bmatrix}
                C & T\\
                0 & 1\\
            \end{bmatrix}
        \end{equation*}
    """
    from_coord = numpy.asarray(from_coord)
    to_coord = numpy.asarray(to_coord)
    from_cen = centroid(from_coord, axis=-1)[:, None]
    to_cen = centroid(to_coord, axis=-1)[:, None]
    from_coord_centered = from_coord - from_cen
    to_coord_centered = to_coord - to_cen

    H = numpy.dot(from_coord_centered, to_coord_centered.T)
    # H = U.dot(numpy.diag(s)).dot(Vt)
    U, s, Vt = numpy.linalg.svd(H, full_matrices=False)
    C = Vt.T.dot(U.T)

    # if numpy.linalg.det(C) < 0:
    #    Vt[2, :] *= -1
    #    C = Vt.T.dot(U.T)

    active = numpy.identity(3, dtype=C.dtype)
    active[0:2, 0:2] = C

    # The formula should be:
    # active[0:2, 2] = (to_cen - C.dot(from_cen)).flatten()
    # But this seems to be more precise:
    active[0:2, 2] = centroid(to_coord - C.dot(from_coord), axis=-1)

    return active


def calc_similarity(
    from_coord: Sequence[numpy.ndarray], to_coord: Sequence[numpy.ndarray]
) -> numpy.ndarray:
    r"""Coordinate array shape is (ndim, ncoord).

    Find the similarity transformation between two sets of coordinates by solving
    this system of equations:

    .. math::

        \begin{align*}
            x' &= a.x - b.y + t_0\\
            y' &= b.x + a.y + t_1\\
            \mathrm{\mathrm{sol}} &= [a, b, t_0, t_1]
        \end{align*}

    .. math::

        \begin{equation*}
            \begin{bmatrix}
                x_1 & -y1 &  1 &  0 \\
                y1 &  x_1 &  0 &  1 \\
                x_2 & -y_2 &  1 &  0 \\
                y_2 &  x_2 &  0 &  1 \\
                \vdots
            \end{bmatrix}.\mathrm{sol}=
            \begin{bmatrix}
                x_1' \\
                y1' \\
                x_2' \\
                y_2' \\
                \vdots
            \end{bmatrix}
        \end{equation*}

    The resulting transformation matrix is

    .. math::

        \begin{equation*}
            \mathrm{sol}=\begin{bmatrix}
                a & -b & t_0\\
                b & a & t_1\\
                0 & 0 & 1
            \end{bmatrix}
        \end{equation*}
    """

    N = len(from_coord[0])

    A = numpy.zeros((2 * N, 4))
    A[::2, 0] = from_coord[0]
    A[1::2, 0] = from_coord[1]
    A[::2, 1] = -from_coord[1]
    A[1::2, 1] = from_coord[0]
    A[::2, 2] = 1
    A[1::2, 3] = 1

    b = numpy.zeros((2 * N, 1))
    b[::2, 0] = to_coord[0]
    b[1::2, 0] = to_coord[1]

    sol = lstsq(A, b)
    active = numpy.identity(3, dtype=sol.dtype)
    active[0:2, 0:2] = [[sol[0], -sol[1]], [sol[1], sol[0]]]
    active[0:2, 2] = sol[2:].flatten()
    return active


def calc_affine(
    from_coord: Sequence[numpy.ndarray], to_coord: Sequence[numpy.ndarray]
) -> numpy.ndarray:
    r"""Coordinate array shape is (ndim, ncoord).

    Find the affine transformation between two sets of coordinates by solving
    this system of equations:

    .. math::

        \begin{align*}
            x' &= a.x + b.y + t_0\\
            y' &= c.x + d.y + t_1\\
            \mathrm{sol} &= [a, b, t_0, c, d, t_1]
        \end{align*}

    .. math::

        \begin{equation*}
            \begin{bmatrix}
                x_1 & y1 & 1 & 0 & 0 & 0 \\
                0 & 0 & 0 & x_1 & y1 & 1 \\
                x_2 & y_2 & 1 & 0 & 0 & 0 \\
                0 & 0 & 0 & x_2 & y_2 & 1 \\
                \vdots
            \end{bmatrix}. \mathrm{sol} =
            \begin{bmatrix}
                x_1' \\
                y1' \\
                x_2' \\
                y_2' \\
                \vdots
            \end{bmatrix}
        \end{equation*}

    The resulting transformation matrix is

    .. math::

        \begin{equation*}
            \mathrm{sol}=\begin{bmatrix}
                a & b & t_0\\
                c & d & t_1\\
                0 & 0 & 1
            \end{bmatrix}
        \end{equation*}
    """

    N = len(from_coord[0])

    A = numpy.zeros((2 * N, 6))
    A[::2, 0] = from_coord[0]
    A[::2, 1] = from_coord[1]
    A[::2, 2] = 1
    A[1::2, 3] = from_coord[0]
    A[1::2, 4] = from_coord[1]
    A[1::2, 5] = 1

    b = numpy.zeros((2 * N, 1))
    b[::2, 0] = to_coord[0]
    b[1::2, 0] = to_coord[1]

    sol = lstsq(A, b)
    active = numpy.identity(3, dtype=sol.dtype)
    active[0:2, :] = sol.reshape((2, 3))
    return active


def calc_projective(
    from_coord: Sequence[numpy.ndarray], to_coord: Sequence[numpy.ndarray]
) -> numpy.ndarray:
    r"""Coordinate array shape is (ndim, ncoord).

    Find the projective transformation between two sets of coordinates by solving
    this system of equations:

    .. math::

        \begin{align*}
            x' &= \frac{a.x + b.y + t_0}{p_x.x+p_y.y+1}
            y' &= \frac{c.x + d.y + t_1}{p_x.x+p_y.y+1}
            x' &= a.x + b.y + t_0 - p_x.x.x' - p_y.y.x'
            y' &= c.x + d.y + t_1 - p_x.x.y' - p_y.y.y'
            \mathrm{sol} &= [a, b, t_0, c, d, t_1 , p_x, p_y]
        \end{align*}

    .. math::

        \begin{equation*}
            \begin{bmatrix}
                x_1 & y1 & 1 & 0 & 0 & 0 & -x_1.x_1' & -y1.x_1'
                0  & 0 & 0 & x_1 & y1 & 1 & -x_1.y1' & -y1.y1'
                x_2 & y_2 & 1 & 0 & 0 & 0 & -x_2.x_2' & -y_2.x_2'
                0 & 0 & 0 & x_2 & y_2 & 1 & -x_2.y_2' & -y_2.y_2'
                \vdots
            \end{bmatrix}. \mathrm{sol} =
            \begin{bmatrix}
                x_1' \\
                y1' \\
                x_2' \\
                y_2' \\
                \vdots
            \end{bmatrix}
        \end{equation*}


    The resulting transformation matrix is

    .. math::

        \begin{equation*}
            \mathrm{sol}=\begin{bmatrix}
                a & b & t_0\\
                c & d & t_1\\
                p_x & p_y & 1
            \end{bmatrix}
        \end{equation*}

    """

    N = len(from_coord[0])

    A = numpy.zeros((2 * N, 8))
    A[::2, 0] = from_coord[0]
    A[::2, 1] = from_coord[1]
    A[::2, 2] = 1
    A[1::2, 3] = from_coord[0]
    A[1::2, 4] = from_coord[1]
    A[1::2, 5] = 1
    A[::2, 6] = -from_coord[0] * to_coord[0]
    A[1::2, 6] = -from_coord[0] * to_coord[1]
    A[::2, 7] = -from_coord[1] * to_coord[0]
    A[1::2, 7] = -from_coord[1] * to_coord[1]

    b = numpy.zeros((2 * N, 1))
    b[::2, 0] = to_coord[0]
    b[1::2, 0] = to_coord[1]

    sol = lstsq(A, b)
    return numpy.append(sol, 1).reshape((3, 3))


def centroid(arr: Sequence[numpy.ndarray], axis: int = 0) -> numpy.ndarray:
    try:
        return trim_mean(arr, 0.1, axis=axis)
    except ValueError:
        return numpy.median(arr, axis=axis)


def lstsq(A: numpy.ndarray, b: numpy.ndarray) -> numpy.ndarray:
    return numpy.linalg.lstsq(A, b, rcond=-1)[0].flatten()



from typing import Tuple
import numpy
from ..detection.base import Features
from ...registry import Registered

__all__ = ["FeatureMatching"]


class FeatureMatching(Registered, register=False):
    def match(
        self,
        from_features: Features,
        to_features: Features,
        from_image: numpy.ndarray,
        to_image: numpy.ndarray,
    ) -> Tuple[Features, Features]:
        raise NotImplementedError



from typing import List
from ewoksorange.bindings import ows_to_ewoks
from ewokscore import execute_graph

import numpy as np
import pytest

try:
    import SimpleITK as sitk
except ImportError:
    sitk = None

try:
    from importlib.resources import files as resource_files
except ImportError:
    from importlib_resources import files as resource_files


@pytest.mark.skipif(
    sitk is None, reason="Workflow uses SimpleITK, which is not available"
)
def test_rigid_workflow_without_qt(tmpdir):
    from ewoksndreg.tests.workflow_tests import workflows

    filename = resource_files(workflows) / "sitk.ows"
    assert_workflow_without_qt(filename, tmpdir)


@pytest.mark.skipif(
    sitk is None, reason="Workflow uses SimpleITK, which is not available"
)
def test_rigid_workflow_with_qt(ewoks_orange_canvas, tmpdir):
    from ewoksndreg.tests.workflow_tests import workflows

    filename = resource_files(workflows) / "sitk.ows"
    assert_workflow_with_qt(ewoks_orange_canvas, filename, tmpdir)


def assert_workflow_without_qt(filename, tmpdir):
    """Execute workflow after converting it to an ewoks workflow"""
    graph = ows_to_ewoks(filename)
    outputs = execute_graph(
        graph, inputs=get_inputs(tmpdir), outputs=[{"all": True}], merge_outputs=False
    )
    for actual, expected in zip(
        outputs["2"]["transformations"], outputs["3"]["transformations"]
    ):
        np.testing.assert_allclose(
            actual.passive, expected.passive, rtol=0.1, atol=0.05
        )
    for actual, expected in zip(outputs["1"]["imagestack"], outputs["0"]["imagestack"]):
        expected = expected[1 - np.isnan(actual)]
        actual = actual[1 - np.isnan(actual)]
        np.testing.assert_allclose(actual, expected, rtol=0.1, atol=0.05)


def assert_workflow_with_qt(ewoks_orange_canvas, filename, tmpdir):
    """Execute workflow using the Qt widgets and signals"""
    ewoks_orange_canvas.load_graph(str(filename), inputs=get_inputs(tmpdir))
    ewoks_orange_canvas.start_workflow()
    ewoks_orange_canvas.wait_widgets(timeout=10)
    outputs = dict(ewoks_orange_canvas.iter_output_values())
    for actual, expected in zip(
        outputs["2"]["transformations"], outputs["3"]["transformations"]
    ):
        np.testing.assert_allclose(
            actual.passive, expected.passive, rtol=0.1, atol=0.05
        )
    for actual, expected in zip(outputs["1"]["imagestack"], outputs["0"]["imagestack"]):
        expected = expected[1 - np.isnan(actual)]
        actual = actual[1 - np.isnan(actual)]
        np.testing.assert_allclose(actual, expected, rtol=0.1, atol=0.05)


def get_inputs(tmpdir) -> List[dict]:
    return []



import numpy
from silx.io import get_data
from ..io import input_stack
from ..io import output_stack


def test_stack_io_numpy():
    values = numpy.zeros((2, 3))
    expected = [values, values + 1, values + 2]

    data = list()
    with output_stack.OutputStackNumpy(data) as stacko:
        stacko.add_point(values)
        stacko.add_points([values + 1, values + 2])
    numpy.testing.assert_array_equal(expected, data)

    with output_stack.output_context() as stacko:
        stacko.add_point(values)
        stacko.add_points([values + 1, values + 2])
        result = stacko.data
    numpy.testing.assert_array_equal(expected, result)

    with input_stack.InputStackNumpy(data) as stacki:
        result = numpy.stack(stacki, axis=0)
    numpy.testing.assert_array_equal(expected, result)

    with input_stack.input_context(data) as stacki:
        result = numpy.stack(stacki, axis=0)
    numpy.testing.assert_array_equal(expected, result)

    with input_stack.InputStackNumpy([data, data], inputs_are_stacks=True) as stacki:
        result = numpy.stack(stacki, axis=0)
    numpy.testing.assert_array_equal(expected + expected, result)

    with input_stack.input_context([data, data], inputs_are_stacks=True) as stacki:
        result = numpy.stack(stacki, axis=0)
    numpy.testing.assert_array_equal(expected + expected, result)


def test_stack_io_hdf5(tmpdir):
    file_path = str(tmpdir / "data1.h5")
    data_path = "entry/instrument/detector/data"
    uri = f"silx://{file_path}::{data_path}"

    values = numpy.zeros((2, 3))
    expected = [values, values + 1, values + 2]

    with output_stack.OutputStackHdf5(uri) as stacko:
        stacko.add_point(values)
        stacko.add_points([values + 1, values + 2])
    numpy.testing.assert_array_equal(expected, get_data(uri))

    file_path = str(tmpdir / "data2.h5")
    uri = f"silx://{file_path}::{data_path}"

    with output_stack.output_context(uri) as stacko:
        stacko.add_point(values)
        stacko.add_points([values + 1, values + 2])
    numpy.testing.assert_array_equal(expected, get_data(uri))

    with input_stack.InputStackHdf5(uri) as stacki:
        data = numpy.stack(list(stacki), axis=0)
    numpy.testing.assert_array_equal(expected, data)

    with input_stack.input_context(uri) as stacki:
        data = numpy.stack(list(stacki), axis=0)
    numpy.testing.assert_array_equal(expected, data)

    with input_stack.InputStackHdf5([uri, uri], inputs_are_stacks=True) as stacki:
        data = numpy.stack(list(stacki), axis=0)
    numpy.testing.assert_array_equal(expected + expected, data)

    with input_stack.input_context([uri, uri], inputs_are_stacks=True) as stacki:
        data = numpy.stack(list(stacki), axis=0)
    numpy.testing.assert_array_equal(expected + expected, data)



import numpy
from ..transformation.homography import type_from_matrix
from ..transformation.types import TransformationType


def test_type_from_matrix():
    matrix = numpy.identity(3)
    assert type_from_matrix(matrix) == TransformationType.identity

    matrix[0:2, 2] = [1, 2]
    assert type_from_matrix(matrix) == TransformationType.translation

    matrix[0, 0] = numpy.cos(0.1)
    matrix[1, 0] = numpy.sin(0.1)
    matrix[0, 1] = -numpy.sin(0.1)
    matrix[1, 1] = numpy.cos(0.1)
    assert type_from_matrix(matrix) == TransformationType.rigid

    matrix[0, 0] *= 1.1
    matrix[1, 1] *= 1.1
    assert type_from_matrix(matrix) == TransformationType.similarity

    matrix[1, 1] *= 1.1
    assert type_from_matrix(matrix) == TransformationType.affine

    matrix[2, 0] = 0.1
    assert type_from_matrix(matrix) == TransformationType.projective



import numpy
from .base import IntensityMapping
from ..math import center
from ..math import fft
from ..transformation import TransformationType
from ..transformation.numpy_backend import NumpyHomography

__all__ = ["NumpyCrossCorrelationIntensityMapping"]


class NumpyCrossCorrelationIntensityMapping(
    IntensityMapping,
    registry_id=IntensityMapping.RegistryId("CrossCorrelation", "Numpy"),
):
    SUPPORTED_TRANSFORMATIONS = ["translation"]

    def __init__(self, transfo_type: TransformationType, **kw) -> None:
        self._to_image_ft = None
        super().__init__(transfo_type, **kw)

    def calculate(
        self, from_image: numpy.ndarray, to_image: numpy.ndarray
    ) -> NumpyHomography:
        from_image_ft = fft.fft2(from_image)
        if to_image is not self._to_image_ft:
            self._to_image_fft = fft.fft2(to_image)
        passive = numpy.identity(3)
        passive[0:2, 2] = determine_shift(from_image_ft, self._to_image_fft)
        return NumpyHomography(passive, self._type)

    def identity(self, dimension: int = 2) -> NumpyHomography:
        return NumpyHomography(
            numpy.identity(dimension + 1), TransformationType.identity
        )


def determine_shift(
    img1ft: numpy.ndarray,
    img2ft: numpy.ndarray,
    sampling: int = 1,
    maxmethod: str = "max",
):
    """Determine shift between images using their Fourier transforms"""
    is1D = img1ft.size in img1ft.shape

    # Calculate shift without subpixel precision
    # Ideally this should be a delta function: real(F.G*/(|F|.|G|))
    # In reality this doesn't seem to work, somehow |F.G*| is better, no idea why
    image_product = img1ft * img2ft.conj()
    # image_product /= ftmodulus(image_product)
    cross_correlation = fft.ifft(image_product)
    shift = cc_maximum(cross_correlation)

    # Shift indices = [0,...,imax,imin,...,-1]
    if is1D:
        _, imax = fft.fft_freqind(cross_correlation.size)
        if shift > imax:
            shift -= cross_correlation.size
    else:
        s = numpy.array(cross_correlation.shape)
        _, imax = fft.fft_freqind(s)
        shift[shift > imax] -= s[shift > imax]

    if sampling <= 1:
        return shift

    # Calculate shift with subpixel precision by interpolating
    # the cross-correlation around the maximum (or center-of-mass).

    # real-space: n, d=1, shift=s
    # super-space: n*sampling, d=1, shift=s*sampling
    if is1D:
        ROIsize = 4 * sampling
    else:
        # ROI in super-space = 3x3 pixels in real-space
        ROIsize = sampling * numpy.array((3, 3))
    _, ksampled = fft.fft_freqind(ROIsize)  # ROI center

    # ROI = [0,1,...,ROIsize-1] - ksampled + shift*sampling  (so that maximum in the middle)
    ROIoffset = ksampled - shift * sampling
    cross_correlation = fft.ifft_interpolate(image_product, ROIoffset, ROIsize)
    # cross_correlation /= img2ft.shape[0]*img2ft.shape[1] * self.sampling ** 2

    # Get fit, maximum, centroid, ...
    shiftsampled = cc_maximum(cross_correlation, maxmethod=maxmethod)

    # Shift from super space to real space
    shift = (shiftsampled - ROIoffset) / sampling

    return shift


def cc_maximum(cross_correlation: numpy.ndarray, maxmethod: str = "max"):
    data = ftmodulus(cross_correlation)
    # data = cross_correlation.real

    if maxmethod == "centroid":
        shift = center.fcentroid(data)
    elif maxmethod == "fit":
        shift = center.fgaussmax(data)
    else:
        shift = center.fmax(data)

    return shift


def ftmodulus(imgft: numpy.ndarray) -> numpy.ndarray:
    imgabs = numpy.abs(imgft)
    imgabs[imgabs < 1.0e-20] = 1
    return imgabs



from typing import Tuple, Optional, Mapping, Dict
from numbers import Number
import numpy
from .base import FeatureMatching
from ..features import Features
from ..features import NumpyKeypointFeatures

__all__ = ["NumpyKeypointFeatureMatching"]


class NumpyKeypointFeatureMatching(
    FeatureMatching,
    registry_id=FeatureMatching.RegistryId("Keypoint", "Numpy"),
    register=False,
):
    def __init__(self, match_options: Optional[Mapping] = None, **kw) -> None:
        if match_options is None:
            match_options = dict()
        match_options.setdefault("npoints", 8)
        self._match_options = match_options
        super().__init__(**kw)

    def match(
        self,
        from_features: Features,
        to_features: Features,
        from_image: numpy.ndarray,
        to_image: numpy.ndarray,
    ) -> Tuple[Features, Features]:
        numpy_from_features = from_features.as_type(NumpyKeypointFeatures)
        numpy_to_features = to_features.as_type(NumpyKeypointFeatures)
        idx_from, idx_to = match_keypoints(
            numpy_from_features,
            numpy_to_features,
            from_image,
            to_image,
            **self._match_options,
        ).T
        return (from_features[idx_from], to_features[idx_to])


def match_keypoints(
    features1: NumpyKeypointFeatures,
    features2: NumpyKeypointFeatures,
    image1: numpy.ndarray,
    image2: numpy.ndarray,
    window_ext=5,
    gauss_sigma=3,
    npoints=None,
    metric="nssd",
) -> Tuple[numpy.ndarray, numpy.ndarray]:
    """Match features based differences between ROI's around
    their coordinates.

    This method returns two arrays:

     - an array of matching indices in `features1` and `features2`
     - an array of matching measures (smaller value means a better match)

    These two arrays are sorted by increasing matching measure.

    The following difference measures are supported:
        - sum of squared differences
    """

    # TODO: add more measures
    # https://towardsdatascience.com/measuring-similarity-in-two-images-using-python-b72233eb53c6
    if image1.shape != image2.shape:
        raise ValueError("'image1' and 'image2' shape do not match")
    wmax = 2 * window_ext + 1
    if image1.shape[0] < wmax or image1.shape[1] < wmax:
        raise ValueError("'window_ext' too large for image shape")

    start1, stop1 = _extract_rois(features1.coordinates, image1.shape, window_ext)
    stack1 = numpy.stack(
        [
            image1[idx_start[0] : idx_stop[0], idx_start[1] : idx_stop[1]]
            for idx_start, idx_stop in zip(start1.T, stop1.T)
        ],
        axis=0,
    )

    start2, stop2 = _extract_rois(features2.coordinates, image2.shape, window_ext)
    stack2 = numpy.stack(
        [
            image1[idx_start[0] : idx_stop[0], idx_start[1] : idx_stop[1]]
            for idx_start, idx_stop in zip(start2.T, stop2.T)
        ],
        axis=0,
    )

    if gauss_sigma:
        weights = _gaussian_weights(window_ext, gauss_sigma)
    else:
        weights = numpy.zeros(wmax, wmax)
    try:
        fmetric = _METRICS[metric]
    except KeyError:
        raise ValueError(f"Unknown image similarity metric '{metric}'")
    dmeas = numpy.vstack(
        [[fmetric(roi1, roi2, weights) for roi2 in stack2] for roi1 in stack1]
    )

    match0 = _matching_indices(dmeas, ref_axis=0)
    match1 = _matching_indices(dmeas, ref_axis=1)
    avgmeas0 = sum(match0.values()) / len(match0)
    avgmeas1 = sum(match1.values()) / len(match1)
    totmeas = avgmeas0 + avgmeas1
    if totmeas:
        w0 = avgmeas0 / totmeas
        w1 = avgmeas1 / totmeas
    else:
        w0 = 0.5
        w1 = 0.5
    match = {
        idx: w0 * match0[idx] + w1 * match1[idx] for idx in match0 if idx in match1
    }
    indices, _ = zip(*sorted(match.items(), key=lambda item: item[1]))
    if npoints:
        indices = indices[:npoints]
    return numpy.array(indices)


def _normalize(image: numpy.ndarray) -> numpy.ndarray:
    mi = numpy.nanmin(image)
    ma = numpy.nanmax(image)
    if mi == ma:
        return image
    return (image - mi) / (ma - mi)


def _gaussian_weights(window_ext: int, sigma=1):
    y, x = numpy.mgrid[-window_ext : window_ext + 1, -window_ext : window_ext + 1]
    g = numpy.zeros(y.shape, dtype=numpy.double)
    g[:] = numpy.exp(-0.5 * (x**2 / sigma**2 + y**2 / sigma**2))
    g /= 2 * numpy.pi * sigma * sigma
    return g


def _extract_rois(
    coordinates: numpy.ndarray, shape: Tuple[int, int], window_ext: int
) -> Tuple[numpy.ndarray, numpy.ndarray]:
    shape = numpy.array(shape).reshape((2, 1))
    coordinates = numpy.round(coordinates).astype(int)
    kpstart = coordinates - window_ext
    kpstop = coordinates + window_ext + 1

    wmax = 2 * window_ext + 1

    for dim, idx in enumerate(kpstart < 0):
        kpstart[dim, idx] = 0
        kpstop[dim, idx] = wmax

    for dim, idx in enumerate(kpstop > shape):
        kpstart[dim, idx] = shape[dim] - wmax
        kpstop[dim, idx] = shape[dim]

    return kpstart, kpstop


def _nssd_metric(
    image1: numpy.ndarray, image2: numpy.ndarray, weights: numpy.ndarray
) -> numpy.ndarray:
    return (weights * (_normalize(image1) - _normalize(image2)) ** 2).sum()


_METRICS = {"nssd": _nssd_metric}


def _matching_indices(
    dmeas: numpy.ndarray, ref_axis=0
) -> Dict[Tuple[Number, Number], Number]:
    """Matching means smallest similarity metric.

    :param dmeas: 2D array with similarity metric (number >= 0) between 2 sets of items
    :returns: map matching index pairs to the corresponding similarity metric
    """
    match_axis = 1 - ref_axis
    ref_idx = range(dmeas.shape[ref_axis])
    match_idx = numpy.argmin(dmeas, axis=match_axis)
    if ref_axis == 0:
        idx0 = ref_idx
        idx1 = match_idx
    else:
        idx0 = match_idx
        idx1 = ref_idx
    return {(i0, i1): dmeas[i0, i1] for i0, i1 in zip(idx0, idx1)}



import numpy
from .base import Features


class SciKitImageDescriptorFeatures(
    Features, registry_id=Features.RegistryId("Descriptor", "SciKitImage")
):
    def __init__(self, keypoints: numpy.ndarray, descriptors: numpy.ndarray) -> None:
        nfeatures, ndim = keypoints.shape
        if ndim != 2:
            raise ValueError("only 2D features are supported")
        self._descriptors = descriptors  # nfeatures x descriptor_size
        self._keypoints = keypoints
        self._coordinates = keypoints.T
        self._nfeatures = nfeatures
        super().__init__()

    def __getitem__(self, idx):
        return type(self)(self._keypoints[idx], self._descriptors[idx])

    @property
    def skimage_keypoints(self) -> numpy.ndarray:
        return self._keypoints

    @property
    def skimage_descriptors(self) -> numpy.ndarray:
        return self._descriptors

    @property
    def coordinates(self) -> numpy.ndarray:
        return self._coordinates

    @property
    def ndim(self) -> int:
        return 2

    @property
    def nfeatures(self) -> int:
        return self._nfeatures



import numpy as np
from skimage.filters import gaussian
from typing import Sequence, Tuple
from ..transformation.homography import Homography
from ..math.normalization import range_normalization


def noisy_eval(stack: np.ndarray) -> float:
    """
    Determines how noisy the images in the given stack are
    This is determined by denoising the image with a gaussian filter and comparing the variance before and after.
    Return value is big if the denoising decreased the variance a lot.

    param stack: array of shape [N,H,W]
    """
    total = 0
    for image in stack:
        image = range_normalization(image)
        var = image.var()
        newvar = gaussian(image).var()
        total += var / newvar
    return total / len(stack)


def peak_eval(stack: np.ndarray, reference: int) -> float:
    """
    Calculate a measure of how reliable a stack of images is to determine the transformations between the images.

    The measure is based on phase cross correlation which generates an image with the peak with coordinates corresponding to the shift.
    This eval calculates how distinguished this peak is by comparing it to the mean of the phase correlation image

    param stack: the aligned or unaligned stack with pixel values in [0,1]
    """
    total = 0
    ref = np.fft.fft2(stack[reference])
    for image in stack:
        transform = np.fft.fft2(image)
        prod = ref * transform.conj()
        peak = np.abs(np.fft.ifft2(prod / np.abs(prod)))
        total += peak.mean() - peak.max()
    return total / len(stack)


def mse_eval(stack: np.ndarray, reference: int) -> float:
    """
    Evaluate success based on the remaining mse error after alignment

    param stack: Aligned stack of images with pixel values in [0,1]
    """
    total = 0
    for image in stack:
        total += np.nanmean((stack[reference] - image) ** 2)
    return total / len(stack)


def smoothness_eval(
    transformations: Sequence[Homography], img_size: Tuple[int, ...]
) -> float:
    """
    Evaluates the transformations by smoothness.

    This is done by transforming the coordinates of the corners of the image and looking at how much these change by successive transformations

    param transformations: Sequence of Homographies
    param img_size: Shape of the images that these transformations are meant for
    """
    i, j = img_size
    points = np.array([[0, 0, i, i], [0, j, 0, j]])
    corners_after = [transfo.apply_coordinates(points) for transfo in transformations]
    corners_diff = [
        (corners_after[i] - corners_after[i - 1]) for i in range(1, len(corners_after))
    ]
    size = np.array([[i], [j]])
    corners_diff = [c / size for c in corners_diff]
    corners_diff = np.linalg.norm(corners_diff, axis=1)
    return np.mean(corners_diff)



import numpy
from ..transformation.numpy_backend import homography_transform_coordinates
from ..transformation.numpy_backend import homography_transform_bounding_box
from ..transformation.numpy_backend import homography_transform_data


def test_homography_transform_coordinates():
    from_coordinates = numpy.random.uniform(-10, 10, (2, 10))

    active1 = numpy.array([[0, 1, 3], [1, 0, -2], [0, 0, 1]])  # transpose + translate
    expected1 = from_coordinates.copy()
    expected1 = expected1[::-1, ...]  # transpose
    expected1 += numpy.array([[3], [-2]])  # translate

    active2 = numpy.array([[-1, 0, -1], [0, -1, 3], [0, 0, 1]])  # inverse + translate
    expected2 = from_coordinates.copy()
    expected2 = -expected2  # inverse
    expected2 += numpy.array([[-1], [3]])  # translate

    active = numpy.stack([active1, active2], axis=0)
    expected = numpy.stack([expected1, expected2], axis=0)

    to_coordinates = homography_transform_coordinates(active1, from_coordinates)
    numpy.testing.assert_allclose(to_coordinates, expected1)

    to_coordinates = homography_transform_coordinates(active2, from_coordinates)
    numpy.testing.assert_allclose(to_coordinates, expected2)

    to_coordinates = homography_transform_coordinates(active, from_coordinates)
    numpy.testing.assert_allclose(to_coordinates, expected)


def test_homography_transform_bounding_box():
    shape = (10, 20)
    active1 = numpy.array([[0, 1, 3], [1, 0, -2], [0, 0, 1]])  # transpose + translate
    expected_xmin1 = numpy.array([3, -2])
    expected_xmax1 = numpy.array([20 + 3, 10 - 2])

    active2 = numpy.array([[-1, 0, -1], [0, -1, 3], [0, 0, 1]])  # inverse + translate
    expected_xmin2 = numpy.array([-10 - 1, -20 + 3])
    expected_xmax2 = expected_xmin2 + numpy.array([10, 20])

    active = numpy.stack([active1, active2], axis=0)
    expected_xmin = numpy.stack([expected_xmin1, expected_xmin2], axis=0)
    expected_xmax = numpy.stack([expected_xmax1, expected_xmax2], axis=0)

    to_xmin, to_xmax = homography_transform_bounding_box(active1, shape)
    numpy.testing.assert_allclose(to_xmin, expected_xmin1)
    numpy.testing.assert_allclose(to_xmax, expected_xmax1)

    to_xmin, to_xmax = homography_transform_bounding_box(active2, shape)
    numpy.testing.assert_allclose(to_xmin, expected_xmin2)
    numpy.testing.assert_allclose(to_xmax, expected_xmax2)

    to_xmin, to_xmax = homography_transform_bounding_box(active, shape)
    numpy.testing.assert_allclose(to_xmin, expected_xmin)
    numpy.testing.assert_allclose(to_xmax, expected_xmax)


def test_homography_transform_data():
    from_img = numpy.zeros((10, 20))
    from_img[5, 3] = 1

    active1 = numpy.array([[0, 1, 5], [1, 0, 8], [0, 0, 1]])  # transpose + translate
    expected1 = numpy.zeros((10, 20))
    expected1[3 + 5, 5 + 8] = 1

    active2 = numpy.array([[-1, 0, 8], [0, -1, 10], [0, 0, 1]])  # inverse + translate
    expected2 = numpy.zeros((10, 20))
    expected2[-5 + 8, -3 + 10] = 1

    passive1 = numpy.linalg.inv(active1)
    passive2 = numpy.linalg.inv(active2)

    to_img = homography_transform_data(passive1, from_img, cval=0)
    numpy.testing.assert_allclose(to_img, expected1)

    to_img = homography_transform_data(passive2, from_img, cval=0)
    numpy.testing.assert_allclose(to_img, expected2)



import numpy
import pytest
from ...math import fft


# http://dsp.stackexchange.com/questions/633/what-data-should-i-use-to-test-an-fft-implementation-and-what-accuracy-should-i
# http://citeseer.ist.psu.edu/viewdoc/summary?doi=10.1.1.38.3924


@pytest.mark.parametrize("convention", [fft.fftConvention.numpy, fft.fftConvention.idl])
@pytest.mark.parametrize("dx", [1, 0.5])
@pytest.mark.parametrize("dy", [1, 0.5])
@pytest.mark.parametrize("nx", list(range(4, 6)))
def test_dirac_delta(convention, dx, dy, nx):
    sig = numpy.zeros(nx)
    sig[0] = 1
    ftsig1 = fft.fft(sig, dx=dx, normconvention=convention)
    ftsig2 = numpy.ones(nx, dtype=complex)
    if convention == fft.fftConvention.idl:
        ftsig2 /= nx
    numpy.testing.assert_allclose(ftsig1, ftsig2)

    ftsig = numpy.ones(nx)
    sig1 = fft.ifft1(ftsig, dx=dx, normconvention=convention)
    sig2 = numpy.zeros(nx, dtype=complex)
    sig2[0] = 1
    if convention == fft.fftConvention.idl:
        sig2[0] = nx
    numpy.testing.assert_allclose(sig1, sig2, atol=1e-10)

    for ny in range(4, 6):
        sig = numpy.zeros((ny, nx))
        sig[0, 0] = 1
        ftsig1 = fft.fft(sig, dx=dx, dy=dy, normconvention=convention)
        ftsig2 = numpy.ones((ny, nx), dtype=complex)
        if convention == fft.fftConvention.idl:
            ftsig2 /= nx * ny
        numpy.testing.assert_allclose(ftsig1, ftsig2)

        ftsig = numpy.ones((ny, nx))
        sig1 = fft.ifft(ftsig, dx=dx, dy=dy, normconvention=convention)
        sig2 = numpy.zeros((ny, nx), dtype=complex)
        sig2[0, 0] = 1
        if convention == fft.fftConvention.idl:
            sig2[0, 0] = nx * ny
        numpy.testing.assert_allclose(sig1, sig2, atol=1e-10)


@pytest.mark.parametrize("convention", [fft.fftConvention.numpy, fft.fftConvention.idl])
@pytest.mark.parametrize("dx", [1, 0.5])
@pytest.mark.parametrize("dy", [1, 0.5])
@pytest.mark.parametrize("nx", list(range(4, 6)))
def test_consistent(convention, dx, dy, nx):
    sig = numpy.random.rand(nx)
    numpy.testing.assert_allclose(
        sig,
        fft.ifft1(
            fft.fft(sig, dx=dx, normconvention=convention),
            dx=dx,
            normconvention=convention,
        ),
    )
    numpy.testing.assert_allclose(
        sig,
        fft.fft(
            fft.ifft1(sig, dx=dx, normconvention=convention),
            dx=dx,
            normconvention=convention,
        ),
    )
    numpy.testing.assert_allclose(
        fft.fft(sig, dx=dx), fft.fft(sig, dx=dx, u=fft.fftfreq(nx, dx))
    )
    for ny in range(4, 6):
        sig = numpy.random.rand(ny, nx)
        numpy.testing.assert_allclose(
            sig,
            fft.ifft(
                fft.fft(sig, dx=dx, dy=dy, normconvention=convention),
                dx=dx,
                dy=dy,
                normconvention=convention,
            ),
        )
        numpy.testing.assert_allclose(
            sig,
            fft.fft(
                fft.ifft(sig, dx=dx, dy=dy, normconvention=convention),
                dx=dx,
                dy=dy,
                normconvention=convention,
            ),
        )
        numpy.testing.assert_allclose(
            fft.fft(sig, dx=dx, dy=dy),
            fft.fft(
                sig,
                dx=dx,
                dy=dy,
                u=fft.fftfreq(nx, dx),
                v=fft.fftfreq(ny, dy),
            ),
        )


@pytest.mark.parametrize("dx", [1, 0.5])
@pytest.mark.parametrize("nx", list(range(4, 6)))
def test_linear(dx, nx):
    x = numpy.random.rand(nx)
    y = numpy.random.rand(nx)
    a = numpy.random.rand(1)[0]
    b = numpy.random.rand(1)[0]
    numpy.testing.assert_allclose(
        fft.fft(a * x + b * y, dx=dx),
        a * fft.fft(x, dx=dx) + b * fft.fft(y, dx=dx),
    )
    numpy.testing.assert_allclose(
        fft.ifft1(a * x + b * y, dx=dx),
        a * fft.ifft1(x, dx=dx) + b * fft.ifft1(y, dx=dx),
    )
    for ny in range(4, 6):
        x = numpy.random.rand(ny, nx)
        y = numpy.random.rand(ny, nx)
        a = numpy.random.rand(1)[0]
        b = numpy.random.rand(1)[0]
        numpy.testing.assert_allclose(
            fft.fft(a * x + b * y, dx=dx),
            a * fft.fft(x, dx=dx) + b * fft.fft(y, dx=dx),
        )
        numpy.testing.assert_allclose(
            fft.ifft(a * x + b * y, dx=dx),
            a * fft.ifft(x, dx=dx) + b * fft.ifft(y, dx=dx),
        )


@pytest.mark.parametrize("nx", list(range(4, 6)))
def test_comp_np(nx):
    sig = numpy.random.rand(nx)
    numpy.testing.assert_allclose(
        numpy.fft.fft(sig), fft.fft(sig, normconvention=fft.fftConvention.numpy)
    )
    numpy.testing.assert_allclose(
        numpy.fft.ifft(sig), fft.ifft1(sig, normconvention=fft.fftConvention.numpy)
    )
    for ny in range(4, 6):
        sig = numpy.random.rand(ny, nx)
        numpy.testing.assert_allclose(
            numpy.fft.fft2(sig),
            fft.fft(sig, normconvention=fft.fftConvention.numpy),
        )
        numpy.testing.assert_allclose(
            numpy.fft.ifft2(sig),
            fft.ifft(sig, normconvention=fft.fftConvention.numpy),
        )

    # Force using non-numpy code
    sig = numpy.random.rand(nx)
    u = fft.fftfreq(nx)
    numpy.testing.assert_allclose(
        numpy.fft.fft(sig), fft.fft(sig, normconvention=fft.fftConvention.numpy, u=u)
    )
    numpy.testing.assert_allclose(
        numpy.fft.ifft(sig),
        fft.ifft(sig, normconvention=fft.fftConvention.numpy, u=u),
    )
    for ny in range(4, 6):
        v = fft.fftfreq(ny)
        sig = numpy.random.rand(ny, nx)
        numpy.testing.assert_allclose(
            numpy.fft.fft2(sig),
            fft.fft(sig, normconvention=fft.fftConvention.numpy, u=u, v=v),
        )
        numpy.testing.assert_allclose(
            numpy.fft.ifft2(sig),
            fft.ifft(sig, normconvention=fft.fftConvention.numpy, u=u, v=v),
        )


@pytest.mark.parametrize("dx", [1, 0.5])
@pytest.mark.parametrize("dy", [1, 0.5])
@pytest.mark.parametrize("nx", list(range(4, 6)))
def test_shift(dx, dy, nx):
    sig1 = numpy.zeros(nx)
    sig1[0] = 1
    sig2 = numpy.zeros(nx)
    ox = 1 + numpy.random.rand(1)[0] * (nx - 2)
    ox = int(ox)
    sig2[ox] = 1
    ftsig1 = fft.fft(sig1, dx=dx)
    ftsig2 = fft.fft(sig2, dx=dx) * numpy.exp(
        2j * numpy.pi * fft.fftfreq(nx, d=dx) * ox * dx
    )
    numpy.testing.assert_allclose(ftsig1, ftsig2)

    for ny in range(4, 6):
        sig1 = numpy.zeros((ny, nx))
        sig1[0, 0] = 1
        sig2 = numpy.zeros((ny, nx))
        ox, oy = 1 + numpy.random.rand(2) * (numpy.array([nx, ny]) - 2)
        ox = int(ox)
        oy = int(oy)
        sig2[oy, ox] = 1
        ftsig1 = fft.fft(sig1, dx=dx, dy=dy)
        ftsig2 = fft.fft(sig2, dx=dx, dy=dy) * numpy.exp(
            2j
            * numpy.pi
            * (
                numpy.add.outer(
                    fft.fftfreq(ny, d=dy) * oy * dy,
                    fft.fftfreq(nx, d=dx) * ox * dx,
                )
            )
        )
        numpy.testing.assert_allclose(ftsig1, ftsig2)


@pytest.mark.parametrize(
    "freqconvention", [fft.fftConvention.numpy, fft.fftConvention.idl]
)
@pytest.mark.parametrize("dx", [1, 0.5])
@pytest.mark.parametrize("dy", [1, 0.5])
@pytest.mark.parametrize("nx", list(range(4, 6)))
def test_freq(freqconvention, dx, dy, nx):
    sig = numpy.random.rand(nx)
    ftsig = fft.fftshift(fft.fft(sig, dx=dx), freqconvention=freqconvention)
    freq = fft.fftshift(
        fft.fftfreq(nx, d=dx, freqconvention=freqconvention),
        freqconvention=freqconvention,
    )
    assert all(numpy.diff(freq) > 0)
    if nx % 2 == 1:
        assert ftsig[0] == numpy.conj(ftsig[-1])

    for ny in range(4, 6):
        sig = numpy.random.rand(ny, nx)
        ftsig = fft.fftshift(
            fft.fft(sig, dx=dx, dy=dy),
            freqconvention=freqconvention,
        )
        u = fft.fftshift(
            fft.fftfreq(nx, d=dx, freqconvention=freqconvention),
            freqconvention=freqconvention,
        )
        v = fft.fftshift(
            fft.fftfreq(ny, d=dy, freqconvention=freqconvention),
            freqconvention=freqconvention,
        )
        assert all(numpy.diff(u) > 0)
        assert all(numpy.diff(v) > 0)
        if nx % 2 == 1 and ny % 2 == 1:
            assert numpy.conj(ftsig[-1, -1]) == ftsig[0, 0]


@pytest.mark.parametrize(
    "freqconvention", [fft.fftConvention.numpy, fft.fftConvention.idl]
)
@pytest.mark.parametrize("dx", [1, 0.5])
@pytest.mark.parametrize("dy", [1, 0.5])
@pytest.mark.parametrize("nx", list(range(4, 6)))
def test_subregion(freqconvention, dx, dy, nx):
    sig = numpy.random.rand(nx)
    ftsig = fft.fft(sig, dx=dx)
    # subregion in real space:
    numpy.testing.assert_allclose(
        sig[1 : nx - 1], fft.ifft1(ftsig, dx=dx, x0=1, x1=nx - 2)
    )
    # subregion in Fourier space:
    u = fft.fftshift(
        fft.fftfreq(nx, d=dx, freqconvention=freqconvention),
        freqconvention=freqconvention,
    )
    numpy.testing.assert_allclose(
        fft.fftshift(ftsig, freqconvention=freqconvention)[1 : nx - 1],
        fft.fft(sig, dx=dx, u=u[1 : nx - 1]),
    )
    for ny in range(4, 6):
        sig = numpy.random.rand(ny, nx)
        ftsig = fft.fft(sig, dx=dx, dy=dy)
        # subregion in real space:
        numpy.testing.assert_allclose(
            sig[1 : ny - 1, 1 : nx - 1],
            fft.ifft(
                ftsig,
                dx=dx,
                x0=1,
                x1=nx - 2,
                dy=dy,
                y0=1,
                y1=ny - 2,
            ),
        )
        # subregion in Fourier space:
        u = fft.fftshift(
            fft.fftfreq(nx, d=dx, freqconvention=freqconvention),
            freqconvention=freqconvention,
        )
        v = fft.fftshift(
            fft.fftfreq(ny, d=dy, freqconvention=freqconvention),
            freqconvention=freqconvention,
        )
        numpy.testing.assert_allclose(
            fft.fftshift(ftsig, freqconvention=freqconvention)[1 : ny - 1, 1 : nx - 1],
            fft.fft(sig, dx=dx, u=u[1 : nx - 1], dy=dy, v=v[1 : ny - 1]),
        )


@pytest.mark.parametrize("nx", list(range(4, 6)))
def test_center(nx):
    sig = numpy.random.rand(nx)
    numpy.testing.assert_allclose(
        fft.fftshift(fft.fft(sig)), fft.fft(sig, centered=True)
    )
    numpy.testing.assert_allclose(
        fft.fftshift(fft.fftfreq(nx)), fft.fftfreq(nx, centered=True)
    )
    numpy.testing.assert_allclose(
        sig, fft.ifft1(fft.fft(sig, centered=True), centered=True)
    )
    for ny in range(4, 6):
        sig = numpy.random.rand(ny, nx)
        numpy.testing.assert_allclose(
            fft.fftshift(fft.fft(sig)), fft.fft(sig, centered=True)
        )
        numpy.testing.assert_allclose(
            sig,
            fft.ifft(fft.fft(sig, centered=True), centered=True),
        )



from typing import Optional, Sequence, List
from ..io.input_stack import InputStack
from ..io.output_stack import OutputStack
from .base import Transformation
from ..math.crop import calc_NaN_edges
import numpy


def apply_transformations(
    input_stack: InputStack,
    output_stack: OutputStack,
    transformations: List[Transformation],
    include: Optional[Sequence[int]] = None,
    cval: int = numpy.nan,
    crop: bool = False,
    interpolation_order: int = 1,
):
    if include is None:
        include = range(len(transformations))
    else:
        if len(include) != len(transformations):
            raise ValueError(
                "Number of transformations and number of items to transform must be the same"
            )

    # calculate slices for cropping
    only_translations = all(
        [x.type in ["identity", "translation"] for x in transformations]
    )
    img_shape = input_stack[0].shape
    low = (0, 0)
    high = (img_shape[0], img_shape[1])
    if crop:
        if only_translations:
            for i in include:
                transformed = transformations[i].apply_data(
                    input_stack[i],
                    offset=None,
                    shape=None,
                    cval=cval,
                    interpolation_order=interpolation_order,
                )
                new_low, new_high = calc_NaN_edges(transformed)
                low = [max(l1, l2) for l1, l2 in zip(low, new_low)]
                high = [min(h1, h2) for h1, h2 in zip(high, new_high)]

        else:
            raise ValueError(
                "Can't meaningfully crop based on transformations that are not translations"
            )

    for i in include:
        transformed = transformations[i].apply_data(
            input_stack[i],
            offset=None,
            shape=None,
            cval=cval,
            interpolation_order=interpolation_order,
        )
        transformed = transformed[low[0] : high[0], low[1] : high[1]]

        output_stack.add_point(transformed)



import pytest


from ewoksndreg.tests.data import data_for_registration

try:
    from ewoksndreg.intensities.torch_metrics import nmi_loss
    import torch
except ImportError:
    torch = None
import numpy as np


@pytest.mark.skipif(torch is None, reason="python version not supported by torch")
def test_rand_nmi():
    rand = torch.rand((3, 1, 20, 20))
    rand2 = torch.rand((3, 1, 20, 20))
    mixed = torch.clone(rand2)
    mixed[..., :10] = rand[..., :10]
    same = nmi_loss(rand, rand)
    half = nmi_loss(rand, mixed)
    random = nmi_loss(rand, rand2)
    torch.testing.assert_close(same < half, torch.ones((3), dtype=torch.bool))
    torch.testing.assert_close(half < random, torch.ones((3), dtype=torch.bool))


@pytest.mark.skipif(torch is None, reason="python version not supported by torch")
def test_img_nmi():
    images, active, passive = data_for_registration.images("translation", nimages=8)
    images = torch.tensor(np.array(images))
    images = images[:, None, ...]

    losses = [nmi_loss(images[0:1], images[i : i + 1]) for i in range(len(images - 1))]
    assert all(losses[i] < losses[i + 1] for i in range(len(losses) - 1))



from ..detection.base import Features
from ...transformation.base import Transformation
from ...registry import Registered


__all__ = ["FeatureMapping"]


class FeatureMapping(Registered, register=False):
    def identity(self) -> Transformation:
        raise NotImplementedError

    def calculate(
        self, from_features: Features, to_features: Features
    ) -> Transformation:
        raise NotImplementedError



from typing import Optional, Dict, Callable, Sequence
import numpy
from skimage.measure import ransac

from .base import FeatureMapping
from ..detection.base import Features
from ...transformation import TransformationType
from ...transformation.numpy_backend import NumpyHomography
from ...transformation.scikitimage_backend import ShiftTransform
from ...transformation.scikitimage_backend import EuclideanTransform
from ...transformation.scikitimage_backend import SimilarityTransform
from ...transformation.scikitimage_backend import AffineTransform
from ...transformation.scikitimage_backend import ProjectiveTransform

__all__ = ["SciKitImageRansacFeatureMapping"]


class SciKitImageRansacFeatureMapping(
    FeatureMapping, registry_id=FeatureMapping.RegistryId("Ransac", "SciKitImage")
):
    def __init__(
        self, transfo_type: TransformationType, solve_options: Optional[Dict] = None
    ) -> None:
        if solve_options is None:
            solve_options = dict()
        solve_options.setdefault("min_samples", 8)
        solve_options.setdefault("residual_threshold", 2)
        solve_options.setdefault("max_trials", 100)
        self._transfo_type = TransformationType(transfo_type)
        self._solver = get_ransac_solver(self._transfo_type, **solve_options)
        super().__init__()

    def calculate(
        self, from_features: Features, to_features: Features
    ) -> NumpyHomography:
        passive = self._solver(to_features.coordinates, from_features.coordinates)
        return NumpyHomography(passive, self._transfo_type)

    def identity(self) -> NumpyHomography:
        return NumpyHomography(numpy.identity(3), TransformationType.identity)


def get_ransac_solver(
    transfo_type: TransformationType, **solve_options
) -> Callable[[Sequence[numpy.ndarray], Sequence[numpy.ndarray]], numpy.ndarray]:
    transfo_type = TransformationType(transfo_type)
    if transfo_type == transfo_type.translation:
        model_classobject = ShiftTransform
    elif transfo_type == transfo_type.rigid:
        model_classobject = EuclideanTransform
    elif transfo_type == transfo_type.similarity:
        model_classobject = SimilarityTransform
    elif transfo_type == transfo_type.affine:
        model_classobject = AffineTransform
    elif transfo_type == transfo_type.projective:
        model_classobject = ProjectiveTransform
    else:
        raise ValueError(f"No ransac solver found for {transfo_type}")

    def ransac_wapper(
        from_coord: Sequence[numpy.ndarray], to_coord: Sequence[numpy.ndarray]
    ) -> numpy.ndarray:
        src = numpy.asarray(from_coord).T
        dst = numpy.asarray(to_coord).T
        tr, inliers = ransac((src, dst), model_classobject, **solve_options)
        return tr.params

    return ransac_wapper



from ewokscore.task import Task
from ..io.input_stack import input_context
from ..features import registration

__all__ = ["Reg2DFeatures"]


class Reg2DFeatures(
    Task,
    input_names=["imagestack", "detector", "matcher", "mapper", "transformation_type"],
    optional_input_names=["inputs_are_stacks", "reference"],
    output_names=["transformations", "features", "matches"],
):
    """Use an feature-based registration method to calculate transformations to register the images in the stack."""

    def run(self):
        detector = registration.FeatureDetector.get_subclass(self.inputs.detector)()
        matcher = registration.FeatureMatching.get_subclass(self.inputs.matcher)()
        mapper = registration.FeatureMapping.get_subclass(self.inputs.mapper)(
            self.inputs.transformation_type
        )
        with input_context(
            self.inputs.imagestack,
            inputs_are_stacks=self.get_input_value("inputs_are_stacks", None),
        ) as stack:
            features = registration.detect_features(stack, detector)
            matches = registration.match_features(
                stack,
                features,
                matcher,
                reference=self.get_input_value("reference", 0),
            )
            self.outputs.transformations = registration.transformations_from_features(
                matches, mapper
            )
            self.outputs.features = features
            self.outputs.matches = matches
